import copy
import re
from collections import Counter as CollectionCounter, defaultdict, deque
from collections.abc import Callable, Hashable as CollectionsHashable, Iterable as CollectionsIterable
from typing import (
    TYPE_CHECKING,
    Any,
    Counter,
    DefaultDict,
    Deque,
    Dict,
    ForwardRef,
    FrozenSet,
    Generator,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Annotated, Final

from . import errors as errors_
from .class_validators import Validator, make_generic_validator, prep_validators
from .error_wrappers import ErrorWrapper
from .errors import ConfigError, InvalidDiscriminator, MissingDiscriminator, NoneIsNotAllowedError
from .types import Json, JsonWrapper
from .typing import (
    NoArgAnyCallable,
    convert_generics,
    display_as_type,
    get_args,
    get_origin,
    is_finalvar,
    is_literal_type,
    is_new_type,
    is_none_type,
    is_typeddict,
    is_typeddict_special,
    is_union,
    new_type_supertype,
)
from .utils import (
    PyObjectStr,
    Representation,
    ValueItems,
    get_discriminator_alias_and_values,
    get_unique_discriminator_alias,
    lenient_isinstance,
    lenient_issubclass,
    sequence_like,
    smart_deepcopy,
)
from .validators import constant_validator, dict_validator, find_validators, validate_json

Required: Any = Ellipsis

T = TypeVar('T')


class UndefinedType:
    def __repr__(self) -> str:
        return 'PydanticUndefined'

    def __copy__(self: T) -> T:
        return self

    def __reduce__(self) -> str:
        return 'Undefined'

    def __deepcopy__(self: T, _: Any) -> T:
        return self


Undefined = UndefinedType()

if TYPE_CHECKING:
    from .class_validators import ValidatorsList # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")
    from .config import BaseConfig # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")
    from .error_wrappers import ErrorList # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")
    from .types import ModelOrDc # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")
    from .typing import AbstractSetIntStr, MappingIntStrAny, ReprArgs # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")

    ValidateReturn = Tuple[Optional[Any], Optional[ErrorList]]
    LocStr = Union[Tuple[Union[int, str], ...], str]
    BoolUndefined = Union[bool, UndefinedType]


class FieldInfo(Representation):
    """
    Captures extra information about a field.
    """

    __slots__ = (
        'default',
        'default_factory',
        'alias',
        'alias_priority',
        'title',
        'description',
        'exclude',
        'include',
        'const',
        'gt',
        'ge',
        'lt',
        'le',
        'multiple_of',
        'allow_inf_nan',
        'max_digits',
        'decimal_places',
        'min_items',
        'max_items',
        'unique_items',
        'min_length',
        'max_length',
        'allow_mutation',
        'repr',
        'regex',
        'discriminator',
        'extra',
    )

    # field constraints with the default value, it's also used in update_from_config below
    __field_constraints__ = {
        'min_length': None,
        'max_length': None,
        'regex': None,
        'gt': None,
        'lt': None,
        'ge': None,
        'le': None,
        'multiple_of': None,
        'allow_inf_nan': None,
        'max_digits': None,
        'decimal_places': None,
        'min_items': None,
        'max_items': None,
        'unique_items': None,
        'allow_mutation': True,
    }

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.default = default
        self.default_factory = kwargs.pop('default_factory', None)
        self.alias = kwargs.pop('alias', None)
        self.alias_priority = kwargs.pop('alias_priority', 2 if self.alias is not None else None)
        self.title = kwargs.pop('title', None)
        self.description = kwargs.pop('description', None)
        self.exclude = kwargs.pop('exclude', None)
        self.include = kwargs.pop('include', None)
        self.const = kwargs.pop('const', None)
        self.gt = kwargs.pop('gt', None)
        self.ge = kwargs.pop('ge', None)
        self.lt = kwargs.pop('lt', None)
        self.le = kwargs.pop('le', None)
        self.multiple_of = kwargs.pop('multiple_of', None)
        self.allow_inf_nan = kwargs.pop('allow_inf_nan', None)
        self.max_digits = kwargs.pop('max_digits', None)
        self.decimal_places = kwargs.pop('decimal_places', None)
        self.min_items = kwargs.pop('min_items', None)
        self.max_items = kwargs.pop('max_items', None)
        self.unique_items = kwargs.pop('unique_items', None)
        self.min_length = kwargs.pop('min_length', None)
        self.max_length = kwargs.pop('max_length', None)
        self.allow_mutation = kwargs.pop('allow_mutation', True)
        self.regex = kwargs.pop('regex', None)
        self.discriminator = kwargs.pop('discriminator', None)
        self.repr = kwargs.pop('repr', True)
        self.extra = kwargs

    def __repr_args__(self) -> 'ReprArgs':

        field_defaults_to_hide: Dict[str, Any] = {
            'repr': True,
            **self.__field_constraints__,
        }

        attrs = ((s, getattr(self, s)) for s in self.__slots__)
        return [(a, v) for a, v in attrs if v != field_defaults_to_hide.get(a, None)]

    def get_constraints(self) -> Set[str]:
        """
        Gets the constraints set on the field by comparing the constraint value with its default value

        :return: the constraints set on field_info
        """
        return {attr for attr, default in self.__field_constraints__.items() if getattr(self, attr) != default}

    def update_from_config(self, from_config: Dict[str, Any]) -> None:
        """
        Update this FieldInfo based on a dict from get_field_info, only fields which have not been set are dated.
        """
        for attr_name, value in from_config.items():
            try:
                current_value = getattr(self, attr_name)
            except AttributeError:
                # attr_name is not an attribute of FieldInfo, it should therefore be added to extra
                # (except if extra already has this value!)
                self.extra.setdefault(attr_name, value)
            else:
                if current_value is self.__field_constraints__.get(attr_name, None):
                    setattr(self, attr_name, value)
                elif attr_name == 'exclude':
                    self.exclude = ValueItems.merge(value, current_value)
                elif attr_name == 'include':
                    self.include = ValueItems.merge(value, current_value, intersect=True)

    def _validate(self) -> None:
        if self.default is not Undefined and self.default_factory is not None:
            raise ValueError('cannot specify both default and default_factory')


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny', Any]] = None,
    include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny', Any]] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    allow_inf_nan: Optional[bool] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field, either for the model schema or complex validation. Some arguments
    apply only to number fields (``int``, ``float``, ``Decimal``) and some apply only to ``str``.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclude: exclude this field while dumping.
      Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param include: include this field while dumping.
      Takes same values as the ``include`` and ``exclude`` arguments on the ``.dict`` method.
    :param const: this field is required and *must* take it's default value
    :param gt: only applies to numbers, requires the field to be "greater than". The schema
      will have an ``exclusiveMinimum`` validation keyword
    :param ge: only applies to numbers, requires the field to be "greater than or equal to". The
      schema will have a ``minimum`` validation keyword
    :param lt: only applies to numbers, requires the field to be "less than". The schema
      will have an ``exclusiveMaximum`` validation keyword
    :param le: only applies to numbers, requires the field to be "less than or equal to". The
      schema will have a ``maximum`` validation keyword
    :param multiple_of: only applies to numbers, requires the field to be "a multiple of". The
      schema will have a ``multipleOf`` validation keyword
    :param allow_inf_nan: only applies to numbers, allows the field to be NaN or infinity (+inf or -inf),
        which is a valid Python float. Default True, set to False for compatibility with JSON.
    :param max_digits: only applies to Decimals, requires the field to have a maximum number
      of digits within the decimal. It does not include a zero before the decimal point or trailing decimal zeroes.
    :param decimal_places: only applies to Decimals, requires the field to have at most a number of decimal places
      allowed. It does not include trailing decimal zeroes.
    :param min_items: only applies to lists, requires the field to have a minimum number of
      elements. The schema will have a ``minItems`` validation keyword
    :param max_items: only applies to lists, requires the field to have a maximum number of
      elements. The schema will have a ``maxItems`` validation keyword
    :param unique_items: only applies to lists, requires the field not to have duplicated
      elements. The schema will have a ``uniqueItems`` validation keyword
    :param min_length: only applies to strings, requires the field to have a minimum length. The
      schema will have a ``minLength`` validation keyword
    :param max_length: only applies to strings, requires the field to have a maximum length. The
      schema will have a ``maxLength`` validation keyword
    :param allow_mutation: a boolean which defaults to True. When False, the field raises a TypeError if the field is
      assigned on an instance.  The BaseModel Config must set validate_assignment to True
    :param regex: only applies to strings, requires the field match against a regular expression
      pattern string. The schema will have a ``pattern`` validation keyword
    :param discriminator: only useful with a (discriminated a.k.a. tagged) `Union` of sub models with a common field.
      The `discriminator` is the name of this common field to shorten validation and improve generated schema
    :param repr: show this field in the representation
    :param **extra: any additional keyword arguments will be added as is to the schema
    """
    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        discriminator=discriminator,
        repr=repr,
        **extra,
    )
    field_info._validate()
    return field_info


# used to be an enum but changed to int's for small performance improvement as less access overhead
SHAPE_SINGLETON = 1
SHAPE_LIST = 2
SHAPE_SET = 3
SHAPE_MAPPING = 4
SHAPE_TUPLE = 5
SHAPE_TUPLE_ELLIPSIS = 6
SHAPE_SEQUENCE = 7
SHAPE_FROZENSET = 8
SHAPE_ITERABLE = 9
SHAPE_GENERIC = 10
SHAPE_DEQUE = 11
SHAPE_DICT = 12
SHAPE_DEFAULTDICT = 13
SHAPE_COUNTER = 14
SHAPE_NAME_LOOKUP = {
    SHAPE_LIST: 'List[{}]',
    SHAPE_SET: 'Set[{}]',
    SHAPE_TUPLE_ELLIPSIS: 'Tuple[{}, ...]',
    SHAPE_SEQUENCE: 'Sequence[{}]',
    SHAPE_FROZENSET: 'FrozenSet[{}]',
    SHAPE_ITERABLE: 'Iterable[{}]',
    SHAPE_DEQUE: 'Deque[{}]',
    SHAPE_DICT: 'Dict[{}]',
    SHAPE_DEFAULTDICT: 'DefaultDict[{}]',
    SHAPE_COUNTER: 'Counter[{}]',
}

MAPPING_LIKE_SHAPES: Set[int] = {SHAPE_DEFAULTDICT, SHAPE_DICT, SHAPE_MAPPING, SHAPE_COUNTER}


class ModelField(Representation):
    __slots__ = (
        'type_',
        'outer_type_',
        'annotation',
        'sub_fields',
        'sub_fields_mapping',
        'key_field',
        'validators',
        'pre_validators',
        'post_validators',
        'default',
        'default_factory',
        'required',
        'final',
        'model_config',
        'name',
        'alias',
        'has_alias',
        'field_info',
        'discriminator_key',
        'discriminator_alias',
        'validate_always',
        'allow_none',
        'shape',
        'class_validators',
        'parse_json',
    )

    def __init__(
        self,
        *,
        name: str,
        type_: Type[Any],
        class_validators: Optional[Dict[str, Validator]],
        model_config: Type['BaseConfig'],
        default: Any = None,
        default_factory: Optional[NoArgAnyCallable] = None,
        required: 'BoolUndefined' = Undefined,
        final: bool = False,
        alias: Optional[str] = None,
        field_info: Optional[FieldInfo] = None,
    ) -> None:

        self.name: str = name
        self.has_alias: bool = alias is not None
        self.alias: str = alias if alias is not None else name
        self.annotation = type_
        self.type_: Any = convert_generics(type_)
        self.outer_type_: Any = type_
        self.class_validators = class_validators or {}
        self.default: Any = default
        self.default_factory: Optional[NoArgAnyCallable] = default_factory
        self.required: 'BoolUndefined' = required
        self.final: bool = final
        self.model_config = model_config
        self.field_info: FieldInfo = field_info or FieldInfo(default)
        self.discriminator_key: Optional[str] = self.field_info.discriminator
        self.discriminator_alias: Optional[str] = self.discriminator_key

        self.allow_none: bool = False
        self.validate_always: bool = False
        self.sub_fields: Optional[List[ModelField]] = None
        self.sub_fields_mapping: Optional[Dict[str, 'ModelField']] = None  # used for discriminated union
        self.key_field: Optional[ModelField] = None
        self.validators: 'ValidatorsList' = []
        self.pre_validators: Optional['ValidatorsList'] = None
        self.post_validators: Optional['ValidatorsList'] = None
        self.parse_json: bool = False
        self.shape: int = SHAPE_SINGLETON
        self.model_config.prepare_field(self)
        self.prepare()

    def get_default(self) -> Any:
        return smart_deepcopy(self.default) if self.default_factory is None else self.default_factory()

    @staticmethod
    def _get_field_info(
        field_name: str, annotation: Any, value: Any, config: Type['BaseConfig']
    ) -> Tuple[FieldInfo, Any]:
        """
        Get a FieldInfo from a root typing.Annotated annotation, value, or config default.

        The FieldInfo may be set in typing.Annotated or the value, but not both. If neither contain
        a FieldInfo, a new one will be created using the config.

        :param field_name: name of the field for use in error messages
        :param annotation: a type hint such as `str` or `Annotated[str, Field(..., min_length=5)]`
        :param value: the field's assigned value
        :param config: the model's config object
        :return: the FieldInfo contained in the `annotation`, the value, or a new one from the config.
        """
        field_info_from_config = config.get_field_info(field_name)

        field_info = None
        if get_origin(annotation) is Annotated:
            field_infos = [arg for arg in get_args(annotation)[1:] if isinstance(arg, FieldInfo)]
            if len(field_infos) > 1:
                raise ValueError(f'cannot specify multiple `Annotated` `Field`s for {field_name!r}')
            field_info = next(iter(field_infos), None)
            if field_info is not None:
                field_info = copy.copy(field_info)
                field_info.update_from_config(field_info_from_config)
                if field_info.default not in (Undefined, Required):
                    raise ValueError(f'`Field` default cannot be set in `Annotated` for {field_name!r}')
                if value is not Undefined and value is not Required:
                    # check also `Required` because of `validate_arguments` that sets `...` as default value
                    field_info.default = value

        if isinstance(value, FieldInfo):
            if field_info is not None:
                raise ValueError(f'cannot specify `Annotated` and value `Field`s together for {field_name!r}')
            field_info = value
            field_info.update_from_config(field_info_from_config)
        elif field_info is None:
            field_info = FieldInfo(value, **field_info_from_config)
        value = None if field_info.default_factory is not None else field_info.default
        field_info._validate()
        return field_info, value

    @classmethod
    def infer(
        cls,
        *,
        name: str,
        value: Any,
        annotation: Any,
        class_validators: Optional[Dict[str, Validator]],
        config: Type['BaseConfig'],
    ) -> 'ModelField':
        from .schema import get_annotation_from_field_info # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")

        field_info, value = cls._get_field_info(name, annotation, value, config)
        required: 'BoolUndefined' = Undefined
        if value is Required:
            required = True
            value = None
        elif value is not Undefined:
            required = False
        annotation = get_annotation_from_field_info(annotation, field_info, name, config.validate_assignment)

        return cls(
            name=name,
            type_=annotation,
            alias=field_info.alias,
            class_validators=class_validators,
            default=value,
            default_factory=field_info.default_factory,
            required=required,
            model_config=config,
            field_info=field_info,
        )

    def set_config(self, config: Type['BaseConfig']) -> None:
        self.model_config = config
        info_from_config = config.get_field_info(self.name)
        config.prepare_field(self)
        new_alias = info_from_config.get('alias')
        new_alias_priority = info_from_config.get('alias_priority') or 0
        if new_alias and new_alias_priority >= (self.field_info.alias_priority or 0):
            self.field_info.alias = new_alias
            self.field_info.alias_priority = new_alias_priority
            self.alias = new_alias
        new_exclude = info_from_config.get('exclude')
        if new_exclude is not None:
            self.field_info.exclude = ValueItems.merge(self.field_info.exclude, new_exclude)
        new_include = info_from_config.get('include')
        if new_include is not None:
            self.field_info.include = ValueItems.merge(self.field_info.include, new_include, intersect=True)

    @property
    def alt_alias(self) -> bool:
        return self.name != self.alias

    def prepare(self) -> None:
        """
        Prepare the field but inspecting self.default, self.type_ etc.

        Note: this method is **not** idempotent (because _type_analysis is not idempotent),
        e.g. calling it it multiple times may modify the field and configure it incorrectly.
        """
        self._set_default_and_type()
        if self.type_.__class__ is ForwardRef or self.type_.__class__ is DeferredType:
            # self.type_ is currently a ForwardRef and there's nothing we can do now,
            # user will need to call model.update_forward_refs()
            return

        self._type_analysis()
        if self.required is Undefined:
            self.required = True
        if self.default is Undefined and self.default_factory is None:
            self.default = None
        self.populate_validators()

    def _set_default_and_type(self) -> None:
        """
        Set the default value, infer the type if needed and check if `None` value is valid.
        """
        if self.default_factory is not None:
            if self.type_ is Undefined:
                raise errors_.ConfigError(
                    f'you need to set the type of field {self.name!r} when using `default_factory`'
                )
            return

        default_value = self.get_default()

        if default_value is not None and self.type_ is Undefined:
            self.type_ = default_value.__class__
            self.outer_type_ = self.type_
            self.annotation = self.type_

        if self.type_ is Undefined:
            raise errors_.ConfigError(f'unable to infer type for attribute "{self.name}"')

        if self.required is False and default_value is None:
            self.allow_none = True

    def _type_analysis(self) -> None:  # noqa: C901 (ignore complexity)
        # typing interface is horrible, we have to do some ugly checks
        if lenient_issubclass(self.type_, JsonWrapper):
            self.type_ = self.type_.inner_type
            self.parse_json = True
        elif lenient_issubclass(self.type_, Json):
            self.type_ = Any
            self.parse_json = True
        elif isinstance(self.type_, TypeVar):
            if self.type_.__bound__:
                self.type_ = self.type_.__bound__
            elif self.type_.__constraints__:
                self.type_ = Union[self.type_.__constraints__]
            else:
                self.type_ = Any
        elif is_new_type(self.type_):
            self.type_ = new_type_supertype(self.type_)

        if self.type_ is Any or self.type_ is object:
            if self.required is Undefined:
                self.required = False
            self.allow_none = True
            return
        elif self.type_ is Pattern or self.type_ is re.Pattern:
            # python 3.7 only, Pattern is a typing object but without sub fields
            return
        elif is_literal_type(self.type_):
            return
        elif is_typeddict(self.type_):
            return

        if is_finalvar(self.type_):
            self.final = True

            if self.type_ is Final:
                self.type_ = Any
            else:
                self.type_ = get_args(self.type_)[0]

            self._type_analysis()
            return

        origin = get_origin(self.type_)

        if origin is Annotated or is_typeddict_special(origin):
            self.type_ = get_args(self.type_)[0]
            self._type_analysis()
            return

        if self.discriminator_key is not None and not is_union(origin):
            raise TypeError('`discriminator` can only be used with `Union` type with more than one variant')

        # add extra check for `collections.abc.Hashable` for python 3.10+ where origin is not `None`
        if origin is None or origin is CollectionsHashable:
            # field is not "typing" object eg. Union, Dict, List etc.
            # allow None for virtual superclasses of NoneType, e.g. Hashable
            if isinstance(self.type_, type) and isinstance(None, self.type_):
                self.allow_none = True
            return
        elif origin is Callable:
            return
        elif is_union(origin):
            types_ = []
            for type_ in get_args(self.type_):
                if is_none_type(type_) or type_ is Any or type_ is object:
                    if self.required is Undefined:
                        self.required = False
                    self.allow_none = True
                if is_none_type(type_):
                    continue
                types_.append(type_)

            if len(types_) == 1:
                # Optional[]
                self.type_ = types_[0]
                # this is the one case where the "outer type" isn't just the original type
                self.outer_type_ = self.type_
                # re-run to correctly interpret the new self.type_
                self._type_analysis()
            else:
                self.sub_fields = [self._create_sub_type(t, f'{self.name}_{display_as_type(t)}') for t in types_]

                if self.discriminator_key is not None:
                    self.prepare_discriminated_union_sub_fields()
            return
        elif issubclass(origin, Tuple):  # type: ignore
            # origin == Tuple without item type
            args = get_args(self.type_)
            if not args:  # plain tuple
                self.type_ = Any
                self.shape = SHAPE_TUPLE_ELLIPSIS
            elif len(args) == 2 and args[1] is Ellipsis:  # e.g. Tuple[int, ...]
                self.type_ = args[0]
                self.shape = SHAPE_TUPLE_ELLIPSIS
                self.sub_fields = [self._create_sub_type(args[0], f'{self.name}_0')]
            elif args == ((),):  # Tuple[()] means empty tuple
                self.shape = SHAPE_TUPLE
                self.type_ = Any
                self.sub_fields = []
            else:
                self.shape = SHAPE_TUPLE
                self.sub_fields = [self._create_sub_type(t, f'{self.name}_{i}') for i, t in enumerate(args)]
            return
        elif issubclass(origin, List):
            # Create self validators
            get_validators = getattr(self.type_, '__get_validators__', None)
            if get_validators:
                self.class_validators.update(
                    {f'list_{i}': Validator(validator, pre=True) for i, validator in enumerate(get_validators())}
                )

            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_LIST
        elif issubclass(origin, Set):
            # Create self validators
            get_validators = getattr(self.type_, '__get_validators__', None)
            if get_validators:
                self.class_validators.update(
                    {f'set_{i}': Validator(validator, pre=True) for i, validator in enumerate(get_validators())}
                )

            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_SET
        elif issubclass(origin, FrozenSet):
            # Create self validators
            get_validators = getattr(self.type_, '__get_validators__', None)
            if get_validators:
                self.class_validators.update(
                    {f'frozenset_{i}': Validator(validator, pre=True) for i, validator in enumerate(get_validators())}
                )

            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_FROZENSET
        elif issubclass(origin, Deque):
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_DEQUE
        elif issubclass(origin, Sequence):
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_SEQUENCE
        # priority to most common mapping: dict
        elif origin is dict or origin is Dict:
            self.key_field = self._create_sub_type(get_args(self.type_)[0], 'key_' + self.name, for_keys=True)
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_DICT
        elif issubclass(origin, DefaultDict):
            self.key_field = self._create_sub_type(get_args(self.type_)[0], 'key_' + self.name, for_keys=True)
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_DEFAULTDICT
        elif issubclass(origin, Counter):
            self.key_field = self._create_sub_type(get_args(self.type_)[0], 'key_' + self.name, for_keys=True)
            self.type_ = int
            self.shape = SHAPE_COUNTER
        elif issubclass(origin, Mapping):
            self.key_field = self._create_sub_type(get_args(self.type_)[0], 'key_' + self.name, for_keys=True)
            self.type_ = get_args(self.type_)[1]
            self.shape = SHAPE_MAPPING
        # Equality check as almost everything inherits form Iterable, including str
        # check for Iterable and CollectionsIterable, as it could receive one even when declared with the other
        elif origin in {Iterable, CollectionsIterable}:
            self.type_ = get_args(self.type_)[0]
            self.shape = SHAPE_ITERABLE
            self.sub_fields = [self._create_sub_type(self.type_, f'{self.name}_type')]
        elif issubclass(origin, Type):  # type: ignore
            return
        elif hasattr(origin, '__get_validators__') or self.model_config.arbitrary_types_allowed:
            # Is a Pydantic-compatible generic that handles itself
            # or we have arbitrary_types_allowed = True
            self.shape = SHAPE_GENERIC
            self.sub_fields = [self._create_sub_type(t, f'{self.name}_{i}') for i, t in enumerate(get_args(self.type_))]
            self.type_ = origin
            return
        else:
            raise TypeError(f'Fields of type "{origin}" are not supported.')

        # type_ has been refined eg. as the type of a List and sub_fields needs to be populated
        self.sub_fields = [self._create_sub_type(self.type_, '_' + self.name)]

    def prepare_discriminated_union_sub_fields(self) -> None:
        """
        Prepare the mapping <discriminator key> -> <ModelField> and update `sub_fields`
        Note that this process can be aborted if a `ForwardRef` is encountered
        """
        assert self.discriminator_key is not None

        if self.type_.__class__ is DeferredType:
            return

        assert self.sub_fields is not None
        sub_fields_mapping: Dict[str, 'ModelField'] = {}
        all_aliases: Set[str] = set()

        for sub_field in self.sub_fields:
            t = sub_field.type_
            if t.__class__ is ForwardRef:
                # Stopping everything...will need to call `update_forward_refs`
                return

            alias, discriminator_values = get_discriminator_alias_and_values(t, self.discriminator_key)
            all_aliases.add(alias)
            for discriminator_value in discriminator_values:
                sub_fields_mapping[discriminator_value] = sub_field

        self.sub_fields_mapping = sub_fields_mapping
        self.discriminator_alias = get_unique_discriminator_alias(all_aliases, self.discriminator_key)

    def _create_sub_type(self, type_: Type[Any], name: str, *, for_keys: bool = False) -> 'ModelField':
        if for_keys:
            class_validators = None
        else:
            # validators for sub items should not have `each_item` as we want to check only the first sublevel
            class_validators = {
                k: Validator(
                    func=v.func,
                    pre=v.pre,
                    each_item=False,
                    always=v.always,
                    check_fields=v.check_fields,
                    skip_on_failure=v.skip_on_failure,
                )
                for k, v in self.class_validators.items()
                if v.each_item
            }

        field_info, _ = self._get_field_info(name, type_, None, self.model_config)

        return self.__class__(
            type_=type_,
            name=name,
            class_validators=class_validators,
            model_config=self.model_config,
            field_info=field_info,
        )

    def populate_validators(self) -> None:
        """
        Prepare self.pre_validators, self.validators, and self.post_validators based on self.type_'s  __get_validators__
        and class validators. This method should be idempotent, e.g. it should be safe to call multiple times
        without mis-configuring the field.
        """
        self.validate_always = getattr(self.type_, 'validate_always', False) or any(
            v.always for v in self.class_validators.values()
        )

        class_validators_ = self.class_validators.values()
        if not self.sub_fields or self.shape == SHAPE_GENERIC:
            get_validators = getattr(self.type_, '__get_validators__', None)
            v_funcs = (
                *[v.func for v in class_validators_ if v.each_item and v.pre],
                *(get_validators() if get_validators else list(find_validators(self.type_, self.model_config))),
                *[v.func for v in class_validators_ if v.each_item and not v.pre],
            )
            self.validators = prep_validators(v_funcs)

        self.pre_validators = []
        self.post_validators = []

        if self.field_info and self.field_info.const:
            self.post_validators.append(make_generic_validator(constant_validator))

        if class_validators_:
            self.pre_validators += prep_validators(v.func for v in class_validators_ if not v.each_item and v.pre)
            self.post_validators += prep_validators(v.func for v in class_validators_ if not v.each_item and not v.pre)

        if self.parse_json:
            self.pre_validators.append(make_generic_validator(validate_json))

        self.pre_validators = self.pre_validators or None
        self.post_validators = self.post_validators or None

    def validate(
        self, v: Any, values: Dict[str, Any], *, loc: 'LocStr', cls: Optional['ModelOrDc'] = None
    ) -> 'ValidateReturn':

        assert self.type_.__class__ is not DeferredType

        if self.type_.__class__ is ForwardRef:
            assert cls is not None
            raise ConfigError(
                f'field "{self.name}" not yet prepared so type is still a ForwardRef, '
                f'you might need to call {cls.__name__}.update_forward_refs().'
            )

        errors: Optional['ErrorList']
        if self.pre_validators:
            v, errors = self._apply_validators(v, values, loc, cls, self.pre_validators)
            if errors:
                return v, errors

        if v is None:
            if is_none_type(self.type_):
                # keep validating
                pass
            elif self.allow_none:
                if self.post_validators:
                    return self._apply_validators(v, values, loc, cls, self.post_validators)
                else:
                    return None, None
            else:
                return v, ErrorWrapper(NoneIsNotAllowedError(), loc)

        if self.shape == SHAPE_SINGLETON:
            v, errors = self._validate_singleton(v, values, loc, cls)
        elif self.shape in MAPPING_LIKE_SHAPES:
            v, errors = self._validate_mapping_like(v, values, loc, cls)
        elif self.shape == SHAPE_TUPLE:
            v, errors = self._validate_tuple(v, values, loc, cls)
        elif self.shape == SHAPE_ITERABLE:
            v, errors = self._validate_iterable(v, values, loc, cls)
        elif self.shape == SHAPE_GENERIC:
            v, errors = self._apply_validators(v, values, loc, cls, self.validators)
        else:
            #  sequence, list, set, generator, tuple with ellipsis, frozen set
            v, errors = self._validate_sequence_like(v, values, loc, cls)

        if not errors and self.post_validators:
            v, errors = self._apply_validators(v, values, loc, cls, self.post_validators)
        return v, errors

    def _validate_sequence_like(  # noqa: C901 (ignore complexity)
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        """
        Validate sequence-like containers: lists, tuples, sets and generators
        Note that large if-else blocks are necessary to enable Cython
        optimization, which is why we disable the complexity check above.
        """
        if not sequence_like(v):
            e: errors_.PydanticTypeError
            if self.shape == SHAPE_LIST:
                e = errors_.ListError()
            elif self.shape in (SHAPE_TUPLE, SHAPE_TUPLE_ELLIPSIS):
                e = errors_.TupleError()
            elif self.shape == SHAPE_SET:
                e = errors_.SetError()
            elif self.shape == SHAPE_FROZENSET:
                e = errors_.FrozenSetError()
            else:
                e = errors_.SequenceError()
            return v, ErrorWrapper(e, loc)

        loc = loc if isinstance(loc, tuple) else (loc,)
        result = []
        errors: List[ErrorList] = []
        for i, v_ in enumerate(v):
            v_loc = *loc, i
            r, ee = self._validate_singleton(v_, values, v_loc, cls)
            if ee:
                errors.append(ee)
            else:
                result.append(r)

        if errors:
            return v, errors

        converted: Union[List[Any], Set[Any], FrozenSet[Any], Tuple[Any, ...], Iterator[Any], Deque[Any]] = result

        if self.shape == SHAPE_SET:
            converted = set(result)
        elif self.shape == SHAPE_FROZENSET:
            converted = frozenset(result)
        elif self.shape == SHAPE_TUPLE_ELLIPSIS:
            converted = tuple(result)
        elif self.shape == SHAPE_DEQUE:
            converted = deque(result, maxlen=getattr(v, 'maxlen', None))
        elif self.shape == SHAPE_SEQUENCE:
            if isinstance(v, tuple):
                converted = tuple(result)
            elif isinstance(v, set):
                converted = set(result)
            elif isinstance(v, Generator):
                converted = iter(result)
            elif isinstance(v, deque):
                converted = deque(result, maxlen=getattr(v, 'maxlen', None))
        return converted, None

    def _validate_iterable(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        """
        Validate Iterables.

        This intentionally doesn't validate values to allow infinite generators.
        """

        try:
            iterable = iter(v)
        except TypeError:
            return v, ErrorWrapper(errors_.IterableError(), loc)
        return iterable, None

    def _validate_tuple(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        e: Optional[Exception] = None
        if not sequence_like(v):
            e = errors_.TupleError()
        else:
            actual_length, expected_length = len(v), len(self.sub_fields)  # type: ignore
            if actual_length != expected_length:
                e = errors_.TupleLengthError(actual_length=actual_length, expected_length=expected_length)

        if e:
            return v, ErrorWrapper(e, loc)

        loc = loc if isinstance(loc, tuple) else (loc,)
        result = []
        errors: List[ErrorList] = []
        for i, (v_, field) in enumerate(zip(v, self.sub_fields)):  # type: ignore
            v_loc = *loc, i
            r, ee = field.validate(v_, values, loc=v_loc, cls=cls)
            if ee:
                errors.append(ee)
            else:
                result.append(r)

        if errors:
            return v, errors
        else:
            return tuple(result), None

    def _validate_mapping_like(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        try:
            v_iter = dict_validator(v)
        except TypeError as exc:
            return v, ErrorWrapper(exc, loc)

        loc = loc if isinstance(loc, tuple) else (loc,)
        result, errors = {}, []
        for k, v_ in v_iter.items():
            v_loc = *loc, '__key__'
            key_result, key_errors = self.key_field.validate(k, values, loc=v_loc, cls=cls)  # type: ignore
            if key_errors:
                errors.append(key_errors)
                continue

            v_loc = *loc, k
            value_result, value_errors = self._validate_singleton(v_, values, v_loc, cls)
            if value_errors:
                errors.append(value_errors)
                continue

            result[key_result] = value_result
        if errors:
            return v, errors
        elif self.shape == SHAPE_DICT:
            return result, None
        elif self.shape == SHAPE_DEFAULTDICT:
            return defaultdict(self.type_, result), None
        elif self.shape == SHAPE_COUNTER:
            return CollectionCounter(result), None
        else:
            return self._get_mapping_value(v, result), None

    def _get_mapping_value(self, original: T, converted: Dict[Any, Any]) -> Union[T, Dict[Any, Any]]:
        """
        When type is `Mapping[KT, KV]` (or another unsupported mapping), we try to avoid
        coercing to `dict` unwillingly.
        """
        original_cls = original.__class__

        if original_cls == dict or original_cls == Dict:
            return converted
        elif original_cls in {defaultdict, DefaultDict}:
            return defaultdict(self.type_, converted)
        else:
            try:
                # Counter, OrderedDict, UserDict, ...
                return original_cls(converted)  # type: ignore
            except TypeError:
                raise RuntimeError(f'Could not convert dictionary to {original_cls.__name__!r}') from None

    def _validate_singleton(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        if self.sub_fields:
            if self.discriminator_key is not None:
                return self._validate_discriminated_union(v, values, loc, cls)

            errors = []

            if self.model_config.smart_union and is_union(get_origin(self.type_)):
                # 1st pass: check if the value is an exact instance of one of the Union types
                # (e.g. to avoid coercing a bool into an int)
                for field in self.sub_fields:
                    if v.__class__ is field.outer_type_:
                        return v, None

                # 2nd pass: check if the value is an instance of any subclass of the Union types
                for field in self.sub_fields:
                    # This whole logic will be improved later on to support more complex `isinstance` checks
                    # It will probably be done once a strict mode is added and be something like:
                    # ```
                    #     value, error = field.validate(v, values, strict=True)
                    #     if error is None:
                    #         return value, None
                    # ```
                    try:
                        if isinstance(v, field.outer_type_):
                            return v, None
                    except TypeError:
                        # compound type
                        if lenient_isinstance(v, get_origin(field.outer_type_)):
                            value, error = field.validate(v, values, loc=loc, cls=cls)
                            if not error:
                                return value, None

            # 1st pass by default or 3rd pass with `smart_union` enabled:
            # check if the value can be coerced into one of the Union types
            for field in self.sub_fields:
                value, error = field.validate(v, values, loc=loc, cls=cls)
                if error:
                    errors.append(error)
                else:
                    return value, None
            return v, errors
        else:
            return self._apply_validators(v, values, loc, cls, self.validators)

    def _validate_discriminated_union(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc']
    ) -> 'ValidateReturn':
        assert self.discriminator_key is not None
        assert self.discriminator_alias is not None

        try:
            try:
                discriminator_value = v[self.discriminator_alias]
            except KeyError:
                if self.model_config.allow_population_by_field_name:
                    discriminator_value = v[self.discriminator_key]
                else:
                    raise
        except KeyError:
            return v, ErrorWrapper(MissingDiscriminator(discriminator_key=self.discriminator_key), loc)
        except TypeError:
            try:
                # BaseModel or dataclass
                discriminator_value = getattr(v, self.discriminator_key)
            except (AttributeError, TypeError):
                return v, ErrorWrapper(MissingDiscriminator(discriminator_key=self.discriminator_key), loc)

        if self.sub_fields_mapping is None:
            assert cls is not None
            raise ConfigError(
                f'field "{self.name}" not yet prepared so type is still a ForwardRef, '
                f'you might need to call {cls.__name__}.update_forward_refs().'
            )

        try:
            sub_field = self.sub_fields_mapping[discriminator_value]
        except (KeyError, TypeError):
            # KeyError: `discriminator_value` is not in the dictionary.
            # TypeError: `discriminator_value` is unhashable.
            assert self.sub_fields_mapping is not None
            return v, ErrorWrapper(
                InvalidDiscriminator(
                    discriminator_key=self.discriminator_key,
                    discriminator_value=discriminator_value,
                    allowed_values=list(self.sub_fields_mapping),
                ),
                loc,
            )
        else:
            if not isinstance(loc, tuple):
                loc = (loc,)
            return sub_field.validate(v, values, loc=(*loc, display_as_type(sub_field.type_)), cls=cls)

    def _apply_validators(
        self, v: Any, values: Dict[str, Any], loc: 'LocStr', cls: Optional['ModelOrDc'], validators: 'ValidatorsList'
    ) -> 'ValidateReturn':
        for validator in validators:
            try:
                v = validator(cls, v, values, self, self.model_config)
            except (ValueError, TypeError, AssertionError) as exc:
                return v, ErrorWrapper(exc, loc)
        return v, None

    def is_complex(self) -> bool:
        """
        Whether the field is "complex" eg. env variables should be parsed as JSON.
        """
        from .main import BaseModel # noqa: PLC0415 - codex_reason("gpt5.2", "avoid circular or optional deps in pydantic compatibility layer")

        return (
            self.shape != SHAPE_SINGLETON
            or hasattr(self.type_, '__pydantic_model__')
            or lenient_issubclass(self.type_, (BaseModel, list, set, frozenset, dict))
        )

    def _type_display(self) -> PyObjectStr:
        t = display_as_type(self.type_)

        if self.shape in MAPPING_LIKE_SHAPES:
            t = f'Mapping[{display_as_type(self.key_field.type_)}, {t}]'  # type: ignore
        elif self.shape == SHAPE_TUPLE:
            t = 'Tuple[{}]'.format(', '.join(display_as_type(f.type_) for f in self.sub_fields))  # type: ignore
        elif self.shape == SHAPE_GENERIC:
            assert self.sub_fields
            t = '{}[{}]'.format(
                display_as_type(self.type_), ', '.join(display_as_type(f.type_) for f in self.sub_fields)
            )
        elif self.shape != SHAPE_SINGLETON:
            t = SHAPE_NAME_LOOKUP[self.shape].format(t)

        if self.allow_none and (self.shape != SHAPE_SINGLETON or not self.sub_fields):
            t = f'Optional[{t}]'
        return PyObjectStr(t)

    def __repr_args__(self) -> 'ReprArgs':
        args = [('name', self.name), ('type', self._type_display()), ('required', self.required)]

        if not self.required:
            if self.default_factory is not None:
                args.append(('default_factory', f'<function {self.default_factory.__name__}>'))
            else:
                args.append(('default', self.default))

        if self.alt_alias:
            args.append(('alias', self.alias))
        return args


class ModelPrivateAttr(Representation):
    __slots__ = ('default', 'default_factory')

    def __init__(self, default: Any = Undefined, *, default_factory: Optional[NoArgAnyCallable] = None) -> None:
        self.default = default
        self.default_factory = default_factory

    def get_default(self) -> Any:
        return smart_deepcopy(self.default) if self.default_factory is None else self.default_factory()

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.default, self.default_factory) == (
            other.default,
            other.default_factory,
        )


def PrivateAttr(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
) -> Any:
    """
    Indicates that attribute is only used internally and never mixed with regular fields.

    Types or values of private attrs are not checked by pydantic and it's up to you to keep them relevant.

    Private attrs are stored in model __slots__.

    :param default: the attribute’s default value
    :param default_factory: callable that will be called when a default value is needed for this attribute
      If both `default` and `default_factory` are set, an error is raised.
    """
    if default is not Undefined and default_factory is not None:
        raise ValueError('cannot specify both default and default_factory')

    return ModelPrivateAttr(
        default,
        default_factory=default_factory,
    )


class DeferredType:
    """
    Used to postpone field preparation, while creating recursive generic models.
    """


def is_finalvar_with_default_val(type_: Type[Any], val: Any) -> bool:
    return is_finalvar(type_) and val is not Undefined and not isinstance(val, FieldInfo)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ConnectionMetadata(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'owner': 'str',
        'full_name': 'str',
        'workspace_url': 'str',
        'client_id': 'str'
    }

    attribute_map = {
        'owner': 'owner',
        'full_name': 'full_name',
        'workspace_url': 'workspace_url',
        'client_id': 'client_id'
    }

    def __init__(self, owner=None, full_name=None, workspace_url=None, client_id=None, local_vars_configuration=None):  # noqa: E501
        """ConnectionMetadata - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._owner = None
        self._full_name = None
        self._workspace_url = None
        self._client_id = None
        self.discriminator = None

        if owner is not None:
            self.owner = owner
        if full_name is not None:
            self.full_name = full_name
        if workspace_url is not None:
            self.workspace_url = workspace_url
        if client_id is not None:
            self.client_id = client_id

    @property
    def owner(self):
        """Gets the owner of this ConnectionMetadata.  # noqa: E501

        Owner of the entity  # noqa: E501

        :return: The owner of this ConnectionMetadata.  # noqa: E501
        :rtype: str
        """
        return self._owner

    @owner.setter
    def owner(self, owner):
        """Sets the owner of this ConnectionMetadata.

        Owner of the entity  # noqa: E501

        :param owner: The owner of this ConnectionMetadata.  # noqa: E501
        :type: str
        """

        self._owner = owner

    @property
    def full_name(self):
        """Gets the full_name of this ConnectionMetadata.  # noqa: E501

        Fully qualified name of the entity  # noqa: E501

        :return: The full_name of this ConnectionMetadata.  # noqa: E501
        :rtype: str
        """
        return self._full_name

    @full_name.setter
    def full_name(self, full_name):
        """Sets the full_name of this ConnectionMetadata.

        Fully qualified name of the entity  # noqa: E501

        :param full_name: The full_name of this ConnectionMetadata.  # noqa: E501
        :type: str
        """

        self._full_name = full_name

    @property
    def workspace_url(self):
        """Gets the workspace_url of this ConnectionMetadata.  # noqa: E501

        Workspace URL for the connection  # noqa: E501

        :return: The workspace_url of this ConnectionMetadata.  # noqa: E501
        :rtype: str
        """
        return self._workspace_url

    @workspace_url.setter
    def workspace_url(self, workspace_url):
        """Sets the workspace_url of this ConnectionMetadata.

        Workspace URL for the connection  # noqa: E501

        :param workspace_url: The workspace_url of this ConnectionMetadata.  # noqa: E501
        :type: str
        """

        self._workspace_url = workspace_url

    @property
    def client_id(self):
        """Gets the client_id of this ConnectionMetadata.  # noqa: E501

        OAuth client ID  # noqa: E501

        :return: The client_id of this ConnectionMetadata.  # noqa: E501
        :rtype: str
        """
        return self._client_id

    @client_id.setter
    def client_id(self, client_id):
        """Sets the client_id of this ConnectionMetadata.

        OAuth client ID  # noqa: E501

        :param client_id: The client_id of this ConnectionMetadata.  # noqa: E501
        :type: str
        """

        self._client_id = client_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ConnectionMetadata):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ConnectionMetadata):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class OrganizationConfigurationResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'customer_billing_type': 'str',
        'invite_code': 'str',
        'registration_url': 'str',
        'expires_at': 'datetime',
        'description': 'str',
        'contract_end_date': 'datetime'
    }

    attribute_map = {
        'customer_billing_type': 'customer_billing_type',
        'invite_code': 'invite_code',
        'registration_url': 'registration_url',
        'expires_at': 'expires_at',
        'description': 'description',
        'contract_end_date': 'contract_end_date'
    }

    def __init__(self, customer_billing_type=None, invite_code=None, registration_url=None, expires_at=None, description=None, contract_end_date=None, local_vars_configuration=None):  # noqa: E501
        """OrganizationConfigurationResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._customer_billing_type = None
        self._invite_code = None
        self._registration_url = None
        self._expires_at = None
        self._description = None
        self._contract_end_date = None
        self.discriminator = None

        self.customer_billing_type = customer_billing_type
        self.invite_code = invite_code
        self.registration_url = registration_url
        if expires_at is not None:
            self.expires_at = expires_at
        if description is not None:
            self.description = description
        if contract_end_date is not None:
            self.contract_end_date = contract_end_date

    @property
    def customer_billing_type(self):
        """Gets the customer_billing_type of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The customer_billing_type of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: str
        """
        return self._customer_billing_type

    @customer_billing_type.setter
    def customer_billing_type(self, customer_billing_type):
        """Sets the customer_billing_type of this OrganizationConfigurationResponse.


        :param customer_billing_type: The customer_billing_type of this OrganizationConfigurationResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and customer_billing_type is None:  # noqa: E501
            raise ValueError("Invalid value for `customer_billing_type`, must not be `None`")  # noqa: E501

        self._customer_billing_type = customer_billing_type

    @property
    def invite_code(self):
        """Gets the invite_code of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The invite_code of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: str
        """
        return self._invite_code

    @invite_code.setter
    def invite_code(self, invite_code):
        """Sets the invite_code of this OrganizationConfigurationResponse.


        :param invite_code: The invite_code of this OrganizationConfigurationResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and invite_code is None:  # noqa: E501
            raise ValueError("Invalid value for `invite_code`, must not be `None`")  # noqa: E501

        self._invite_code = invite_code

    @property
    def registration_url(self):
        """Gets the registration_url of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The registration_url of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: str
        """
        return self._registration_url

    @registration_url.setter
    def registration_url(self, registration_url):
        """Sets the registration_url of this OrganizationConfigurationResponse.


        :param registration_url: The registration_url of this OrganizationConfigurationResponse.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and registration_url is None:  # noqa: E501
            raise ValueError("Invalid value for `registration_url`, must not be `None`")  # noqa: E501

        self._registration_url = registration_url

    @property
    def expires_at(self):
        """Gets the expires_at of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The expires_at of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._expires_at

    @expires_at.setter
    def expires_at(self, expires_at):
        """Sets the expires_at of this OrganizationConfigurationResponse.


        :param expires_at: The expires_at of this OrganizationConfigurationResponse.  # noqa: E501
        :type: datetime
        """

        self._expires_at = expires_at

    @property
    def description(self):
        """Gets the description of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The description of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this OrganizationConfigurationResponse.


        :param description: The description of this OrganizationConfigurationResponse.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def contract_end_date(self):
        """Gets the contract_end_date of this OrganizationConfigurationResponse.  # noqa: E501


        :return: The contract_end_date of this OrganizationConfigurationResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._contract_end_date

    @contract_end_date.setter
    def contract_end_date(self, contract_end_date):
        """Sets the contract_end_date of this OrganizationConfigurationResponse.


        :param contract_end_date: The contract_end_date of this OrganizationConfigurationResponse.  # noqa: E501
        :type: datetime
        """

        self._contract_end_date = contract_end_date

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrganizationConfigurationResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrganizationConfigurationResponse):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreditGrantRecordV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'amount_usd': 'float',
        'effective_date_start': 'date',
        'effective_date_end': 'date'
    }

    attribute_map = {
        'amount_usd': 'amount_usd',
        'effective_date_start': 'effective_date_start',
        'effective_date_end': 'effective_date_end'
    }

    def __init__(self, amount_usd=None, effective_date_start=None, effective_date_end=None, local_vars_configuration=None):  # noqa: E501
        """CreditGrantRecordV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._amount_usd = None
        self._effective_date_start = None
        self._effective_date_end = None
        self.discriminator = None

        self.amount_usd = amount_usd
        self.effective_date_start = effective_date_start
        self.effective_date_end = effective_date_end

    @property
    def amount_usd(self):
        """Gets the amount_usd of this CreditGrantRecordV2.  # noqa: E501

        Amount of USD available to use from the credit.  # noqa: E501

        :return: The amount_usd of this CreditGrantRecordV2.  # noqa: E501
        :rtype: float
        """
        return self._amount_usd

    @amount_usd.setter
    def amount_usd(self, amount_usd):
        """Sets the amount_usd of this CreditGrantRecordV2.

        Amount of USD available to use from the credit.  # noqa: E501

        :param amount_usd: The amount_usd of this CreditGrantRecordV2.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and amount_usd is None:  # noqa: E501
            raise ValueError("Invalid value for `amount_usd`, must not be `None`")  # noqa: E501

        self._amount_usd = amount_usd

    @property
    def effective_date_start(self):
        """Gets the effective_date_start of this CreditGrantRecordV2.  # noqa: E501

        UTC date the credit can first be used (inclusive).  # noqa: E501

        :return: The effective_date_start of this CreditGrantRecordV2.  # noqa: E501
        :rtype: date
        """
        return self._effective_date_start

    @effective_date_start.setter
    def effective_date_start(self, effective_date_start):
        """Sets the effective_date_start of this CreditGrantRecordV2.

        UTC date the credit can first be used (inclusive).  # noqa: E501

        :param effective_date_start: The effective_date_start of this CreditGrantRecordV2.  # noqa: E501
        :type: date
        """
        if self.local_vars_configuration.client_side_validation and effective_date_start is None:  # noqa: E501
            raise ValueError("Invalid value for `effective_date_start`, must not be `None`")  # noqa: E501

        self._effective_date_start = effective_date_start

    @property
    def effective_date_end(self):
        """Gets the effective_date_end of this CreditGrantRecordV2.  # noqa: E501

        UTC date the credit can last be used (inclusive).  # noqa: E501

        :return: The effective_date_end of this CreditGrantRecordV2.  # noqa: E501
        :rtype: date
        """
        return self._effective_date_end

    @effective_date_end.setter
    def effective_date_end(self, effective_date_end):
        """Sets the effective_date_end of this CreditGrantRecordV2.

        UTC date the credit can last be used (inclusive).  # noqa: E501

        :param effective_date_end: The effective_date_end of this CreditGrantRecordV2.  # noqa: E501
        :type: date
        """
        if self.local_vars_configuration.client_side_validation and effective_date_end is None:  # noqa: E501
            raise ValueError("Invalid value for `effective_date_end`, must not be `None`")  # noqa: E501

        self._effective_date_end = effective_date_end

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreditGrantRecordV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreditGrantRecordV2):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class OperatorEvent(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'dataset_id': 'str',
        'event_time': 'float',
        'event_type': 'OperatorEventType',
        'message': 'str'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'dataset_id': 'dataset_id',
        'event_time': 'event_time',
        'event_type': 'event_type',
        'message': 'message'
    }

    def __init__(self, id=None, name=None, dataset_id=None, event_time=None, event_type=None, message=None, local_vars_configuration=None):  # noqa: E501
        """OperatorEvent - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._dataset_id = None
        self._event_time = None
        self._event_type = None
        self._message = None
        self.discriminator = None

        self.id = id
        self.name = name
        self.dataset_id = dataset_id
        self.event_time = event_time
        self.event_type = event_type
        self.message = message

    @property
    def id(self):
        """Gets the id of this OperatorEvent.  # noqa: E501


        :return: The id of this OperatorEvent.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OperatorEvent.


        :param id: The id of this OperatorEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this OperatorEvent.  # noqa: E501


        :return: The name of this OperatorEvent.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this OperatorEvent.


        :param name: The name of this OperatorEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def dataset_id(self):
        """Gets the dataset_id of this OperatorEvent.  # noqa: E501


        :return: The dataset_id of this OperatorEvent.  # noqa: E501
        :rtype: str
        """
        return self._dataset_id

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        """Sets the dataset_id of this OperatorEvent.


        :param dataset_id: The dataset_id of this OperatorEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and dataset_id is None:  # noqa: E501
            raise ValueError("Invalid value for `dataset_id`, must not be `None`")  # noqa: E501

        self._dataset_id = dataset_id

    @property
    def event_time(self):
        """Gets the event_time of this OperatorEvent.  # noqa: E501


        :return: The event_time of this OperatorEvent.  # noqa: E501
        :rtype: float
        """
        return self._event_time

    @event_time.setter
    def event_time(self, event_time):
        """Sets the event_time of this OperatorEvent.


        :param event_time: The event_time of this OperatorEvent.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and event_time is None:  # noqa: E501
            raise ValueError("Invalid value for `event_time`, must not be `None`")  # noqa: E501

        self._event_time = event_time

    @property
    def event_type(self):
        """Gets the event_type of this OperatorEvent.  # noqa: E501


        :return: The event_type of this OperatorEvent.  # noqa: E501
        :rtype: OperatorEventType
        """
        return self._event_type

    @event_type.setter
    def event_type(self, event_type):
        """Sets the event_type of this OperatorEvent.


        :param event_type: The event_type of this OperatorEvent.  # noqa: E501
        :type: OperatorEventType
        """
        if self.local_vars_configuration.client_side_validation and event_type is None:  # noqa: E501
            raise ValueError("Invalid value for `event_type`, must not be `None`")  # noqa: E501

        self._event_type = event_type

    @property
    def message(self):
        """Gets the message of this OperatorEvent.  # noqa: E501


        :return: The message of this OperatorEvent.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this OperatorEvent.


        :param message: The message of this OperatorEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and message is None:  # noqa: E501
            raise ValueError("Invalid value for `message`, must not be `None`")  # noqa: E501

        self._message = message

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OperatorEvent):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OperatorEvent):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
from datetime import date
from typing import Optional

from anyscale._private.models import ModelBase


@dataclass(frozen=True)
class DownloadCSVFilters(ModelBase):
    """Filters to use when downloading usage CSVs.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.aggregated_instance_usage.models import DownloadCSVFilters

download_csv_filters = DownloadCSVFilters(
    # Start date (UTC inclusive) for the usage CSV.
    start_date="2024-10-01",
    # End date (UTC inclusive) for the usage CSV.
    end_date="2024-10-31",
    # Optional cloud name to filter by.
    cloud="cloud_name",
    # Optional project name to filter by.
    project="project_name",
    # Optional directory to save the CSV to.
    directory="/directory",
    # Optional hide progress bar.
    hide_progress_bar=False,
)
"""

    def _validate_date(self, date_string: str):
        if not isinstance(date_string, str):
            raise TypeError("date must be a string.")
        try:
            date.fromisoformat(date_string)
        except ValueError:
            raise ValueError("Incorrect date format, should be YYYY-MM-DD")

    start_date: str = field(
        metadata={"docstring": "Start date (UTC inclusive) for the usage CSV."}
    )

    def _validate_start_date(self, start_date: str):
        self._validate_date(start_date)

    end_date: str = field(
        metadata={"docstring": "End date (UTC inclusive) for the usage CSV."}
    )

    def _validate_end_date(self, end_date: str):
        self._validate_date(end_date)

    cloud: Optional[str] = field(
        default=None, metadata={"docstring": "Optional cloud name to filter by."},
    )

    def _validate_cloud(self, cloud: Optional[str]):
        if cloud is not None and not isinstance(cloud, str):
            raise TypeError("cloud must be a string.")

    project: Optional[str] = field(
        default=None, metadata={"docstring": "Optional project name to filter by."},
    )

    def _validate_project(self, project: Optional[str]):
        if project is not None and not isinstance(project, str):
            raise TypeError("project must be a string.")

    directory: Optional[str] = field(
        default=None, metadata={"docstring": "Optional directory to save the CSV to."},
    )

    def _validate_directory(self, directory: Optional[str]):
        if directory is not None and not isinstance(directory, str):
            raise TypeError("directory must be a string.")

    hide_progress_bar: Optional[bool] = field(
        default=False, metadata={"docstring": "Optional hide progress bar."},
    )

    def _validate_hide_progress_bar(self, hide_progress_bar: Optional[bool]):
        if hide_progress_bar is not None and not isinstance(hide_progress_bar, bool):
            raise TypeError("hide_progress_bar must be a boolean.")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class GetOrCreateBuildFromImageUriRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'image_uri': 'str',
        'ray_version': 'str',
        'registry_login_secret': 'str',
        'cluster_env_name': 'str'
    }

    attribute_map = {
        'image_uri': 'image_uri',
        'ray_version': 'ray_version',
        'registry_login_secret': 'registry_login_secret',
        'cluster_env_name': 'cluster_env_name'
    }

    def __init__(self, image_uri=None, ray_version=None, registry_login_secret=None, cluster_env_name=None, local_vars_configuration=None):  # noqa: E501
        """GetOrCreateBuildFromImageUriRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._image_uri = None
        self._ray_version = None
        self._registry_login_secret = None
        self._cluster_env_name = None
        self.discriminator = None

        self.image_uri = image_uri
        if ray_version is not None:
            self.ray_version = ray_version
        if registry_login_secret is not None:
            self.registry_login_secret = registry_login_secret
        if cluster_env_name is not None:
            self.cluster_env_name = cluster_env_name

    @property
    def image_uri(self):
        """Gets the image_uri of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501

        The image uri to get or create a build from.  # noqa: E501

        :return: The image_uri of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :rtype: str
        """
        return self._image_uri

    @image_uri.setter
    def image_uri(self, image_uri):
        """Sets the image_uri of this GetOrCreateBuildFromImageUriRequest.

        The image uri to get or create a build from.  # noqa: E501

        :param image_uri: The image_uri of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and image_uri is None:  # noqa: E501
            raise ValueError("Invalid value for `image_uri`, must not be `None`")  # noqa: E501

        self._image_uri = image_uri

    @property
    def ray_version(self):
        """Gets the ray_version of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501

        The Ray version to use for the build.  # noqa: E501

        :return: The ray_version of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :rtype: str
        """
        return self._ray_version

    @ray_version.setter
    def ray_version(self, ray_version):
        """Sets the ray_version of this GetOrCreateBuildFromImageUriRequest.

        The Ray version to use for the build.  # noqa: E501

        :param ray_version: The ray_version of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :type: str
        """

        self._ray_version = ray_version

    @property
    def registry_login_secret(self):
        """Gets the registry_login_secret of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501

        The registry login secret to use for the build.  # noqa: E501

        :return: The registry_login_secret of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :rtype: str
        """
        return self._registry_login_secret

    @registry_login_secret.setter
    def registry_login_secret(self, registry_login_secret):
        """Sets the registry_login_secret of this GetOrCreateBuildFromImageUriRequest.

        The registry login secret to use for the build.  # noqa: E501

        :param registry_login_secret: The registry_login_secret of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :type: str
        """

        self._registry_login_secret = registry_login_secret

    @property
    def cluster_env_name(self):
        """Gets the cluster_env_name of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501

        The name of the cluster environment to use for the build. Only used if the image uri is from external registry.  # noqa: E501

        :return: The cluster_env_name of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :rtype: str
        """
        return self._cluster_env_name

    @cluster_env_name.setter
    def cluster_env_name(self, cluster_env_name):
        """Sets the cluster_env_name of this GetOrCreateBuildFromImageUriRequest.

        The name of the cluster environment to use for the build. Only used if the image uri is from external registry.  # noqa: E501

        :param cluster_env_name: The cluster_env_name of this GetOrCreateBuildFromImageUriRequest.  # noqa: E501
        :type: str
        """

        self._cluster_env_name = cluster_env_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GetOrCreateBuildFromImageUriRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GetOrCreateBuildFromImageUriRequest):
            return True

        return self.to_dict() != other.to_dict()



from typing import Optional

import click
from rich import print as rprint
import tabulate
import yaml

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand
from anyscale.policy.models import PolicyBinding, PolicyConfig


log = BlockLogger()


@click.group("policy", help="Manage resource permission policies.")
def policy_cli() -> None:
    pass


@policy_cli.command(
    name="set",
    cls=AnyscaleCommand,
    example=command_examples.POLICY_SET_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--resource-type",
    required=True,
    type=click.Choice(["cloud", "project", "organization"], case_sensitive=False),
    help="Resource type ('cloud', 'project', or 'organization').",
)
@click.option(
    "--resource-id",
    required=False,
    default=None,
    type=str,
    help="Resource ID (e.g., cld_abc123, prj_xyz789). Required for 'cloud' and 'project' types, not allowed for 'organization'.",
)
@click.option(
    "-f",
    "--config-file",
    required=True,
    type=click.Path(exists=True),
    help="Path to a YAML config file with policy bindings.",
)
def set_policy(
    resource_type: str, resource_id: Optional[str], config_file: str,
) -> None:
    """
    Set user group permission policy for a resource.

    The config file should be in YAML format with bindings list.

    For organization policies, --resource-id cannot be specified, the policy will
    be set for your current organization automatically.

    Example policy.yaml:

    \b
    bindings:
      - role_name: collaborator
        principals:
          - ug_abc123
      - role_name: readonly
        principals:
          - ug_def456
          - ug_ghi789

    Valid role_name values:

    \b
      Cloud:        collaborator, readonly
      Project:      owner, collaborator, readonly
      Organization: owner, collaborator
    """
    # Validate resource_id based on resource_type
    if resource_type.lower() == "organization":
        if resource_id is not None:
            raise click.ClickException(
                "--resource-id cannot be specified for 'organization' resource type. "
                "The policy will be set for your current organization automatically."
            )
        display_id = "your organization"
    elif resource_id is None:
        raise click.ClickException(
            f"--resource-id is required for resource type '{resource_type}'."
        )
    else:
        display_id = resource_id

    log.info(f"Setting policy for {resource_type} {display_id}...")

    try:
        with open(config_file) as f:
            config_dict = yaml.safe_load(f)
    except yaml.YAMLError as e:
        raise click.ClickException(f"Failed to parse YAML file '{config_file}': {e}")

    if config_dict is None:
        raise click.ClickException(
            f"Invalid config file '{config_file}': file is empty."
        )
    if not isinstance(config_dict, dict):
        raise click.ClickException(
            f"Invalid config file '{config_file}': expected a YAML mapping with top-level 'bindings'."
        )

    try:
        bindings = [
            PolicyBinding(role_name=b["role_name"], principals=b["principals"],)
            for b in config_dict.get("bindings", [])
        ]
        config = PolicyConfig(bindings=bindings)
    except (KeyError, TypeError) as e:
        raise click.ClickException(f"Invalid config file format: {e}")

    try:
        anyscale.policy.set(
            resource_type=resource_type, resource_id=resource_id, config=config,
        )
    except ValueError as e:
        raise click.ClickException(f"Failed to set policy: {e}")

    log.info(f"Policy for {resource_type} {display_id} has been updated.")


@policy_cli.command(
    name="get",
    cls=AnyscaleCommand,
    example=command_examples.POLICY_GET_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--resource-type",
    required=True,
    type=click.Choice(["cloud", "project", "organization"], case_sensitive=False),
    help="Resource type ('cloud', 'project', or 'organization').",
)
@click.option(
    "--resource-id",
    required=False,
    default=None,
    type=str,
    help="Resource ID (e.g., cld_abc123, prj_xyz789). Required for 'cloud' and 'project' types, not allowed for 'organization'.",
)
def get_policy(resource_type: str, resource_id: Optional[str],) -> None:
    """
    Get user group permission policy for a resource.

    For organization policies, --resource-id cannot be specified, the policy for
    your current organization will be returned automatically.
    """
    # Validate resource_id based on resource_type
    if resource_type.lower() == "organization":
        if resource_id is not None:
            raise click.ClickException(
                "--resource-id cannot be specified for 'organization' resource type. "
                "The policy will be retrieved for your current organization automatically."
            )
        display_id = "your organization"
    elif resource_id is None:
        raise click.ClickException(
            f"--resource-id is required for resource type '{resource_type}'."
        )
    else:
        display_id = resource_id

    try:
        policy = anyscale.policy.get(
            resource_type=resource_type, resource_id=resource_id,
        )
    except ValueError as e:
        log.error(f"Failed to get policy: {e}")
        return

    if not policy.bindings:
        log.info(f"No policy bindings found for {resource_type} {display_id}.")
        return

    status_str = policy.sync_status.value

    log.info(f"Policy for {resource_type} {display_id}:")

    table_data = []
    for binding in policy.bindings:
        for principal in binding.principals:
            table_data.append((binding.role_name, principal, status_str))

    table = tabulate.tabulate(
        table_data, headers=["Role", "Principal (User Group ID)", "Process Status"],
    )
    rprint(table)


@policy_cli.command(
    name="list",
    cls=AnyscaleCommand,
    example=command_examples.POLICY_LIST_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--resource-type",
    required=True,
    type=click.Choice(["cloud", "project"], case_sensitive=False),
    help="Resource type to list policies for ('cloud' or 'project').",
)
def list_policies(resource_type: str,) -> None:
    """
    List permission policies for all resources of a specific type.

    Only shows resources that have bindings configured.
    """
    try:
        policies = anyscale.policy.list(resource_type=resource_type)
    except ValueError as e:
        log.error(f"Failed to list policies: {e}")
        return

    if not policies:
        log.info(f"No {resource_type}s found.")
        return

    # Filter to only show policies with bindings
    policies_with_bindings = [p for p in policies if p.bindings]

    if not policies_with_bindings:
        log.info(f"No bindings configured for any {resource_type}s.")
        return

    for policy in policies_with_bindings:
        log.info(f"\n{policy.resource_type}: {policy.resource_id}")

        status_str = policy.sync_status.value
        table_data = []
        for binding in policy.bindings:
            for principal in binding.principals:
                table_data.append((binding.role_name, principal, status_str))
        table = tabulate.tabulate(
            table_data, headers=["Role", "Principal (User Group ID)", "Process Status"],
        )
        rprint(table)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ClusterDashboardNode(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'node_id': 'str',
        'node_ip': 'str',
        'name': 'str',
        'status': 'NodeStatus',
        'is_head_node': 'bool',
        'resources': 'dict(str, float)',
        'labels': 'dict(str, str)',
        'instance_type': 'str',
        'start_time_ms': 'int',
        'end_time_ms': 'int',
        'ray_session_name': 'str'
    }

    attribute_map = {
        'node_id': 'node_id',
        'node_ip': 'node_ip',
        'name': 'name',
        'status': 'status',
        'is_head_node': 'is_head_node',
        'resources': 'resources',
        'labels': 'labels',
        'instance_type': 'instance_type',
        'start_time_ms': 'start_time_ms',
        'end_time_ms': 'end_time_ms',
        'ray_session_name': 'ray_session_name'
    }

    def __init__(self, node_id=None, node_ip=None, name=None, status=None, is_head_node=None, resources=None, labels=None, instance_type=None, start_time_ms=None, end_time_ms=None, ray_session_name=None, local_vars_configuration=None):  # noqa: E501
        """ClusterDashboardNode - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._node_id = None
        self._node_ip = None
        self._name = None
        self._status = None
        self._is_head_node = None
        self._resources = None
        self._labels = None
        self._instance_type = None
        self._start_time_ms = None
        self._end_time_ms = None
        self._ray_session_name = None
        self.discriminator = None

        self.node_id = node_id
        self.node_ip = node_ip
        self.name = name
        self.status = status
        self.is_head_node = is_head_node
        self.resources = resources
        self.labels = labels
        if instance_type is not None:
            self.instance_type = instance_type
        if start_time_ms is not None:
            self.start_time_ms = start_time_ms
        if end_time_ms is not None:
            self.end_time_ms = end_time_ms
        if ray_session_name is not None:
            self.ray_session_name = ray_session_name

    @property
    def node_id(self):
        """Gets the node_id of this ClusterDashboardNode.  # noqa: E501


        :return: The node_id of this ClusterDashboardNode.  # noqa: E501
        :rtype: str
        """
        return self._node_id

    @node_id.setter
    def node_id(self, node_id):
        """Sets the node_id of this ClusterDashboardNode.


        :param node_id: The node_id of this ClusterDashboardNode.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and node_id is None:  # noqa: E501
            raise ValueError("Invalid value for `node_id`, must not be `None`")  # noqa: E501

        self._node_id = node_id

    @property
    def node_ip(self):
        """Gets the node_ip of this ClusterDashboardNode.  # noqa: E501


        :return: The node_ip of this ClusterDashboardNode.  # noqa: E501
        :rtype: str
        """
        return self._node_ip

    @node_ip.setter
    def node_ip(self, node_ip):
        """Sets the node_ip of this ClusterDashboardNode.


        :param node_ip: The node_ip of this ClusterDashboardNode.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and node_ip is None:  # noqa: E501
            raise ValueError("Invalid value for `node_ip`, must not be `None`")  # noqa: E501

        self._node_ip = node_ip

    @property
    def name(self):
        """Gets the name of this ClusterDashboardNode.  # noqa: E501


        :return: The name of this ClusterDashboardNode.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ClusterDashboardNode.


        :param name: The name of this ClusterDashboardNode.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def status(self):
        """Gets the status of this ClusterDashboardNode.  # noqa: E501


        :return: The status of this ClusterDashboardNode.  # noqa: E501
        :rtype: NodeStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ClusterDashboardNode.


        :param status: The status of this ClusterDashboardNode.  # noqa: E501
        :type: NodeStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def is_head_node(self):
        """Gets the is_head_node of this ClusterDashboardNode.  # noqa: E501


        :return: The is_head_node of this ClusterDashboardNode.  # noqa: E501
        :rtype: bool
        """
        return self._is_head_node

    @is_head_node.setter
    def is_head_node(self, is_head_node):
        """Sets the is_head_node of this ClusterDashboardNode.


        :param is_head_node: The is_head_node of this ClusterDashboardNode.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and is_head_node is None:  # noqa: E501
            raise ValueError("Invalid value for `is_head_node`, must not be `None`")  # noqa: E501

        self._is_head_node = is_head_node

    @property
    def resources(self):
        """Gets the resources of this ClusterDashboardNode.  # noqa: E501


        :return: The resources of this ClusterDashboardNode.  # noqa: E501
        :rtype: dict(str, float)
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """Sets the resources of this ClusterDashboardNode.


        :param resources: The resources of this ClusterDashboardNode.  # noqa: E501
        :type: dict(str, float)
        """
        if self.local_vars_configuration.client_side_validation and resources is None:  # noqa: E501
            raise ValueError("Invalid value for `resources`, must not be `None`")  # noqa: E501

        self._resources = resources

    @property
    def labels(self):
        """Gets the labels of this ClusterDashboardNode.  # noqa: E501


        :return: The labels of this ClusterDashboardNode.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._labels

    @labels.setter
    def labels(self, labels):
        """Sets the labels of this ClusterDashboardNode.


        :param labels: The labels of this ClusterDashboardNode.  # noqa: E501
        :type: dict(str, str)
        """
        if self.local_vars_configuration.client_side_validation and labels is None:  # noqa: E501
            raise ValueError("Invalid value for `labels`, must not be `None`")  # noqa: E501

        self._labels = labels

    @property
    def instance_type(self):
        """Gets the instance_type of this ClusterDashboardNode.  # noqa: E501


        :return: The instance_type of this ClusterDashboardNode.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type):
        """Sets the instance_type of this ClusterDashboardNode.


        :param instance_type: The instance_type of this ClusterDashboardNode.  # noqa: E501
        :type: str
        """

        self._instance_type = instance_type

    @property
    def start_time_ms(self):
        """Gets the start_time_ms of this ClusterDashboardNode.  # noqa: E501


        :return: The start_time_ms of this ClusterDashboardNode.  # noqa: E501
        :rtype: int
        """
        return self._start_time_ms

    @start_time_ms.setter
    def start_time_ms(self, start_time_ms):
        """Sets the start_time_ms of this ClusterDashboardNode.


        :param start_time_ms: The start_time_ms of this ClusterDashboardNode.  # noqa: E501
        :type: int
        """

        self._start_time_ms = start_time_ms

    @property
    def end_time_ms(self):
        """Gets the end_time_ms of this ClusterDashboardNode.  # noqa: E501


        :return: The end_time_ms of this ClusterDashboardNode.  # noqa: E501
        :rtype: int
        """
        return self._end_time_ms

    @end_time_ms.setter
    def end_time_ms(self, end_time_ms):
        """Sets the end_time_ms of this ClusterDashboardNode.


        :param end_time_ms: The end_time_ms of this ClusterDashboardNode.  # noqa: E501
        :type: int
        """

        self._end_time_ms = end_time_ms

    @property
    def ray_session_name(self):
        """Gets the ray_session_name of this ClusterDashboardNode.  # noqa: E501


        :return: The ray_session_name of this ClusterDashboardNode.  # noqa: E501
        :rtype: str
        """
        return self._ray_session_name

    @ray_session_name.setter
    def ray_session_name(self, ray_session_name):
        """Sets the ray_session_name of this ClusterDashboardNode.


        :param ray_session_name: The ray_session_name of this ClusterDashboardNode.  # noqa: E501
        :type: str
        """

        self._ray_session_name = ray_session_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ClusterDashboardNode):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ClusterDashboardNode):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ComputeNodeType(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'instance_type': 'str',
        'resources': 'Resources',
        'physical_resources': 'PhysicalResources',
        'required_resources': 'PhysicalResources',
        'labels': 'dict(str, str)',
        'required_labels': 'dict(str, str)',
        'aws_advanced_configurations_json': 'object',
        'gcp_advanced_configurations_json': 'object',
        'advanced_configurations_json': 'object',
        'flags': 'object'
    }

    attribute_map = {
        'name': 'name',
        'instance_type': 'instance_type',
        'resources': 'resources',
        'physical_resources': 'physical_resources',
        'required_resources': 'required_resources',
        'labels': 'labels',
        'required_labels': 'required_labels',
        'aws_advanced_configurations_json': 'aws_advanced_configurations_json',
        'gcp_advanced_configurations_json': 'gcp_advanced_configurations_json',
        'advanced_configurations_json': 'advanced_configurations_json',
        'flags': 'flags'
    }

    def __init__(self, name=None, instance_type=None, resources=None, physical_resources=None, required_resources=None, labels=None, required_labels=None, aws_advanced_configurations_json=None, gcp_advanced_configurations_json=None, advanced_configurations_json=None, flags=None, local_vars_configuration=None):  # noqa: E501
        """ComputeNodeType - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._instance_type = None
        self._resources = None
        self._physical_resources = None
        self._required_resources = None
        self._labels = None
        self._required_labels = None
        self._aws_advanced_configurations_json = None
        self._gcp_advanced_configurations_json = None
        self._advanced_configurations_json = None
        self._flags = None
        self.discriminator = None

        self.name = name
        if instance_type is not None:
            self.instance_type = instance_type
        if resources is not None:
            self.resources = resources
        if physical_resources is not None:
            self.physical_resources = physical_resources
        if required_resources is not None:
            self.required_resources = required_resources
        if labels is not None:
            self.labels = labels
        if required_labels is not None:
            self.required_labels = required_labels
        if aws_advanced_configurations_json is not None:
            self.aws_advanced_configurations_json = aws_advanced_configurations_json
        if gcp_advanced_configurations_json is not None:
            self.gcp_advanced_configurations_json = gcp_advanced_configurations_json
        if advanced_configurations_json is not None:
            self.advanced_configurations_json = advanced_configurations_json
        if flags is not None:
            self.flags = flags

    @property
    def name(self):
        """Gets the name of this ComputeNodeType.  # noqa: E501

        An arbitrary name for this node type, which will be registered with OSS available_node_types.   # noqa: E501

        :return: The name of this ComputeNodeType.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ComputeNodeType.

        An arbitrary name for this node type, which will be registered with OSS available_node_types.   # noqa: E501

        :param name: The name of this ComputeNodeType.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def instance_type(self):
        """Gets the instance_type of this ComputeNodeType.  # noqa: E501

        The cloud provider instance type to use for this node. Optional when using free pod shapes with required_resources.  # noqa: E501

        :return: The instance_type of this ComputeNodeType.  # noqa: E501
        :rtype: str
        """
        return self._instance_type

    @instance_type.setter
    def instance_type(self, instance_type):
        """Sets the instance_type of this ComputeNodeType.

        The cloud provider instance type to use for this node. Optional when using free pod shapes with required_resources.  # noqa: E501

        :param instance_type: The instance_type of this ComputeNodeType.  # noqa: E501
        :type: str
        """

        self._instance_type = instance_type

    @property
    def resources(self):
        """Gets the resources of this ComputeNodeType.  # noqa: E501

        Declaration of node resources for Autoscaler.  # noqa: E501

        :return: The resources of this ComputeNodeType.  # noqa: E501
        :rtype: Resources
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """Sets the resources of this ComputeNodeType.

        Declaration of node resources for Autoscaler.  # noqa: E501

        :param resources: The resources of this ComputeNodeType.  # noqa: E501
        :type: Resources
        """

        self._resources = resources

    @property
    def physical_resources(self):
        """Gets the physical_resources of this ComputeNodeType.  # noqa: E501

        DEPRECATED: This field is deprecated. Please use required_resources instead.  # noqa: E501

        :return: The physical_resources of this ComputeNodeType.  # noqa: E501
        :rtype: PhysicalResources
        """
        return self._physical_resources

    @physical_resources.setter
    def physical_resources(self, physical_resources):
        """Sets the physical_resources of this ComputeNodeType.

        DEPRECATED: This field is deprecated. Please use required_resources instead.  # noqa: E501

        :param physical_resources: The physical_resources of this ComputeNodeType.  # noqa: E501
        :type: PhysicalResources
        """

        self._physical_resources = physical_resources

    @property
    def required_resources(self):
        """Gets the required_resources of this ComputeNodeType.  # noqa: E501

        Required resources for compute node type which specifies the actual CPU, memory, and GPU resources that should be allocated for this node type.  # noqa: E501

        :return: The required_resources of this ComputeNodeType.  # noqa: E501
        :rtype: PhysicalResources
        """
        return self._required_resources

    @required_resources.setter
    def required_resources(self, required_resources):
        """Sets the required_resources of this ComputeNodeType.

        Required resources for compute node type which specifies the actual CPU, memory, and GPU resources that should be allocated for this node type.  # noqa: E501

        :param required_resources: The required_resources of this ComputeNodeType.  # noqa: E501
        :type: PhysicalResources
        """

        self._required_resources = required_resources

    @property
    def labels(self):
        """Gets the labels of this ComputeNodeType.  # noqa: E501

        Labels to associate the node with for scheduling purposes. Defaults to the list of Ray & Anyscale default labels.  # noqa: E501

        :return: The labels of this ComputeNodeType.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._labels

    @labels.setter
    def labels(self, labels):
        """Sets the labels of this ComputeNodeType.

        Labels to associate the node with for scheduling purposes. Defaults to the list of Ray & Anyscale default labels.  # noqa: E501

        :param labels: The labels of this ComputeNodeType.  # noqa: E501
        :type: dict(str, str)
        """

        self._labels = labels

    @property
    def required_labels(self):
        """Gets the required_labels of this ComputeNodeType.  # noqa: E501

        Required labels that must be present on the node for scheduling purposes.  # noqa: E501

        :return: The required_labels of this ComputeNodeType.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._required_labels

    @required_labels.setter
    def required_labels(self, required_labels):
        """Sets the required_labels of this ComputeNodeType.

        Required labels that must be present on the node for scheduling purposes.  # noqa: E501

        :param required_labels: The required_labels of this ComputeNodeType.  # noqa: E501
        :type: dict(str, str)
        """

        self._required_labels = required_labels

    @property
    def aws_advanced_configurations_json(self):
        """Gets the aws_advanced_configurations_json of this ComputeNodeType.  # noqa: E501

        The advanced configuration json that we pass directly AWS APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :return: The aws_advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :rtype: object
        """
        return self._aws_advanced_configurations_json

    @aws_advanced_configurations_json.setter
    def aws_advanced_configurations_json(self, aws_advanced_configurations_json):
        """Sets the aws_advanced_configurations_json of this ComputeNodeType.

        The advanced configuration json that we pass directly AWS APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :param aws_advanced_configurations_json: The aws_advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :type: object
        """

        self._aws_advanced_configurations_json = aws_advanced_configurations_json

    @property
    def gcp_advanced_configurations_json(self):
        """Gets the gcp_advanced_configurations_json of this ComputeNodeType.  # noqa: E501

        The advanced configuration json that we pass directly GCP APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :return: The gcp_advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :rtype: object
        """
        return self._gcp_advanced_configurations_json

    @gcp_advanced_configurations_json.setter
    def gcp_advanced_configurations_json(self, gcp_advanced_configurations_json):
        """Sets the gcp_advanced_configurations_json of this ComputeNodeType.

        The advanced configuration json that we pass directly GCP APIs when launching an instance. We may do some validation on this json and reject the json if it is using a configuration that Anyscale does not support.  # noqa: E501

        :param gcp_advanced_configurations_json: The gcp_advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :type: object
        """

        self._gcp_advanced_configurations_json = gcp_advanced_configurations_json

    @property
    def advanced_configurations_json(self):
        """Gets the advanced_configurations_json of this ComputeNodeType.  # noqa: E501

        Advanced configurations for this compute node type to pass to the cloud provider when launching this instance.  # noqa: E501

        :return: The advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :rtype: object
        """
        return self._advanced_configurations_json

    @advanced_configurations_json.setter
    def advanced_configurations_json(self, advanced_configurations_json):
        """Sets the advanced_configurations_json of this ComputeNodeType.

        Advanced configurations for this compute node type to pass to the cloud provider when launching this instance.  # noqa: E501

        :param advanced_configurations_json: The advanced_configurations_json of this ComputeNodeType.  # noqa: E501
        :type: object
        """

        self._advanced_configurations_json = advanced_configurations_json

    @property
    def flags(self):
        """Gets the flags of this ComputeNodeType.  # noqa: E501

        A set of advanced node-level flags that can be used to configure a particular workload.  # noqa: E501

        :return: The flags of this ComputeNodeType.  # noqa: E501
        :rtype: object
        """
        return self._flags

    @flags.setter
    def flags(self, flags):
        """Sets the flags of this ComputeNodeType.

        A set of advanced node-level flags that can be used to configure a particular workload.  # noqa: E501

        :param flags: The flags of this ComputeNodeType.  # noqa: E501
        :type: object
        """

        self._flags = flags

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ComputeNodeType):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ComputeNodeType):
            return True

        return self.to_dict() != other.to_dict()



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
from enum import Enum


class Bytes(int, Enum):
    """Gets the number of bytes per unit."""

    KB = 1024
    MB = 1024 * KB
    GB = 1024 * MB



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CommitLedgerRecordV2(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'ledger_record_type': 'CreditType',
        'amount_usd': 'float',
        'timestamp': 'datetime',
        'commit_ledger_item_type': 'CommitLedgerItemType'
    }

    attribute_map = {
        'ledger_record_type': 'ledger_record_type',
        'amount_usd': 'amount_usd',
        'timestamp': 'timestamp',
        'commit_ledger_item_type': 'commit_ledger_item_type'
    }

    def __init__(self, ledger_record_type=None, amount_usd=None, timestamp=None, commit_ledger_item_type=None, local_vars_configuration=None):  # noqa: E501
        """CommitLedgerRecordV2 - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._ledger_record_type = None
        self._amount_usd = None
        self._timestamp = None
        self._commit_ledger_item_type = None
        self.discriminator = None

        if ledger_record_type is not None:
            self.ledger_record_type = ledger_record_type
        self.amount_usd = amount_usd
        self.timestamp = timestamp
        self.commit_ledger_item_type = commit_ledger_item_type

    @property
    def ledger_record_type(self):
        """Gets the ledger_record_type of this CommitLedgerRecordV2.  # noqa: E501


        :return: The ledger_record_type of this CommitLedgerRecordV2.  # noqa: E501
        :rtype: CreditType
        """
        return self._ledger_record_type

    @ledger_record_type.setter
    def ledger_record_type(self, ledger_record_type):
        """Sets the ledger_record_type of this CommitLedgerRecordV2.


        :param ledger_record_type: The ledger_record_type of this CommitLedgerRecordV2.  # noqa: E501
        :type: CreditType
        """

        self._ledger_record_type = ledger_record_type

    @property
    def amount_usd(self):
        """Gets the amount_usd of this CommitLedgerRecordV2.  # noqa: E501

        Amount of USD added or removed from the credit.  # noqa: E501

        :return: The amount_usd of this CommitLedgerRecordV2.  # noqa: E501
        :rtype: float
        """
        return self._amount_usd

    @amount_usd.setter
    def amount_usd(self, amount_usd):
        """Sets the amount_usd of this CommitLedgerRecordV2.

        Amount of USD added or removed from the credit.  # noqa: E501

        :param amount_usd: The amount_usd of this CommitLedgerRecordV2.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and amount_usd is None:  # noqa: E501
            raise ValueError("Invalid value for `amount_usd`, must not be `None`")  # noqa: E501

        self._amount_usd = amount_usd

    @property
    def timestamp(self):
        """Gets the timestamp of this CommitLedgerRecordV2.  # noqa: E501

        UTC timestamp of when the record occurred.  # noqa: E501

        :return: The timestamp of this CommitLedgerRecordV2.  # noqa: E501
        :rtype: datetime
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        """Sets the timestamp of this CommitLedgerRecordV2.

        UTC timestamp of when the record occurred.  # noqa: E501

        :param timestamp: The timestamp of this CommitLedgerRecordV2.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and timestamp is None:  # noqa: E501
            raise ValueError("Invalid value for `timestamp`, must not be `None`")  # noqa: E501

        self._timestamp = timestamp

    @property
    def commit_ledger_item_type(self):
        """Gets the commit_ledger_item_type of this CommitLedgerRecordV2.  # noqa: E501

        Type of commit ledger item.  # noqa: E501

        :return: The commit_ledger_item_type of this CommitLedgerRecordV2.  # noqa: E501
        :rtype: CommitLedgerItemType
        """
        return self._commit_ledger_item_type

    @commit_ledger_item_type.setter
    def commit_ledger_item_type(self, commit_ledger_item_type):
        """Sets the commit_ledger_item_type of this CommitLedgerRecordV2.

        Type of commit ledger item.  # noqa: E501

        :param commit_ledger_item_type: The commit_ledger_item_type of this CommitLedgerRecordV2.  # noqa: E501
        :type: CommitLedgerItemType
        """
        if self.local_vars_configuration.client_side_validation and commit_ledger_item_type is None:  # noqa: E501
            raise ValueError("Invalid value for `commit_ledger_item_type`, must not be `None`")  # noqa: E501

        self._commit_ledger_item_type = commit_ledger_item_type

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CommitLedgerRecordV2):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CommitLedgerRecordV2):
            return True

        return self.to_dict() != other.to_dict()



from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Optional

from anyscale._private.models import ModelBase


@dataclass(frozen=True)
class AdminCreateUser(ModelBase):
    """User to be created by an admin.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.user.models import AdminCreateUser

admin_create_user = AdminCreateUser(
    # First name of the user to be created.
    name="name",
    # Email of the user to be created.
    email="test@anyscale.com",
    # Password for the user account being created.
    password="",
    # Whether the user is an SSO user. SSO users can log in using SSO.
    is_sso_user=False,
    # Optional last name of the user to be created.
    lastname="lastname",
    # Optional title of the user to be created.
    title="title",
)
"""
    name: str = field(metadata={"docstring": "First name of the user to be created."})

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

    email: str = field(metadata={"docstring": "Email of the user to be created."})

    def _validate_email(self, email: str):
        if not isinstance(email, str):
            raise TypeError("email must be a string.")

    password: Optional[str] = field(
        metadata={"docstring": "Password for the user account being created."}
    )

    def _validate_password(self, password: Optional[str]):
        if password is not None and not isinstance(password, str):
            raise TypeError("password must be a string.")

    is_sso_user: bool = field(
        metadata={
            "docstring": "Whether the user is an SSO user. SSO users can log in using SSO."
        },
    )

    def _validate_is_sso_user(self, is_sso_user: bool):
        if not isinstance(is_sso_user, bool):
            raise TypeError("is_sso_user must be a boolean.")

    lastname: Optional[str] = field(
        default=None,
        metadata={"docstring": "Optional last name of the user to be created."},
    )

    def _validate_lastname(self, lastname: Optional[str]):
        if lastname is not None and not isinstance(lastname, str):
            raise TypeError("lastname must be a string.")

    title: Optional[str] = field(
        default=None,
        metadata={"docstring": "Optional title of the user to be created."},
    )

    def _validate_title(self, title: Optional[str]):
        if title is not None and not isinstance(title, str):
            raise TypeError("title must be a string.")


@dataclass(frozen=True)
class AdminCreateUsers(ModelBase):
    """Users to be created by an admin.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.user.models import AdminCreateUser

admin_create_user = AdminCreateUser(
    # First name of the user to be created.
    name="name",
    # Email of the user to be created.
    email="test@anyscale.com",
    # Password for the user account being created.
    password="",
    # Whether the user is an SSO user. SSO users can log in using SSO.
    is_sso_user=False,
    # Optional last name of the user to be created.
    lastname="lastname",
    # Optional title of the user to be created.
    title="title",
)
admin_create_users = AdminCreateUsers(
    # Users to be created by an admin.
    create_users=[admin_create_user]
)
"""
    create_users: List[Dict[str, Any]] = field(
        metadata={"docstring": "Users to be created by an admin."}
    )

    def _validate_create_users(self, create_users: List[Dict[str, Any]]):
        if not isinstance(create_users, list):
            raise TypeError("create_users must be a list.")


@dataclass(frozen=True)
class AdminCreatedUser(ModelBase):
    """User account created by an admin that has organization collaborator permissions.
    """

    __doc_py_example__ = """\
import anyscale
from anyscale.user.models import AdminCreatedUser

admin_create_user = AdminCreateUser(
    # First name of the user to be created.
    name="name",
    # Email of the user to be created.
    email="test@anyscale.com",
    # Password for the user account being created.
    password="",
    # Whether the user is an SSO user. SSO users can log in using SSO.
    is_sso_user=False,
    # Optional last name of the user to be created.
    lastname="lastname",
    # Optional title of the user to be created.
    title="title",
)
admin_created_users: List[AdminCreatedUser] = anyscale.user.admin_batch_create([admin_create_user])
"""
    user_id: str = field(
        metadata={"docstring": "ID of the user that has been created."}
    )

    def _validate_user_id(self, user_id: str):
        if not isinstance(user_id, str):
            raise TypeError("user_id must be a string.")

    name: str = field(
        metadata={"docstring": "First name of the user that has been created."}
    )

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

    email: str = field(
        metadata={"docstring": "Email of the user that has been created."}
    )

    def _validate_email(self, email: str):
        if not isinstance(email, str):
            raise TypeError("email must be a string.")

    created_at: datetime = field(
        metadata={"docstring": "The timestamp of when the user is created."}
    )

    def _validate_created_at(self, created_at: datetime):
        if not isinstance(created_at, datetime):
            raise TypeError("created_at must be a datetime.")

    is_sso_user: bool = field(
        metadata={
            "docstring": "Whether the user is an SSO user. SSO users can log in using SSO."
        },
    )

    def _validate_is_sso_user(self, is_sso_user: bool):
        if not isinstance(is_sso_user, bool):
            raise TypeError("is_sso_user must be a boolean.")

    lastname: Optional[str] = field(
        default=None,
        metadata={"docstring": "Optional last name of the user that has been created."},
    )

    def _validate_lastname(self, lastname: Optional[str]):
        if lastname is not None and not isinstance(lastname, str):
            raise TypeError("lastname must be a string.")

    title: Optional[str] = field(
        default=None,
        metadata={"docstring": "Optional title of the user that has been created."},
    )

    def _validate_title(self, title: Optional[str]):
        if title is not None and not isinstance(title, str):
            raise TypeError("title must be a string.")


@dataclass(frozen=True)
class User(ModelBase):
    """Collaborator returned by ``anyscale.user`` APIs."""

    __doc_py_example__ = """\
import anyscale

for user in anyscale.user.list(max_items=5):
    print(f"{user.email} ({user.permission_level})")
"""

    email: str = field(
        metadata={"docstring": "Email address associated with the collaborator."}
    )

    def _validate_email(self, email: str):
        if not isinstance(email, str):
            raise TypeError("email must be a string.")

    name: str = field(metadata={"docstring": "Display name of the collaborator."})

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

    created_at: datetime = field(
        metadata={"docstring": "Timestamp for when the collaborator was created."}
    )

    def _validate_created_at(self, created_at: datetime):
        if not isinstance(created_at, datetime):
            raise TypeError("created_at must be a datetime.")

    permission_level: str = field(
        metadata={"docstring": "Organization permission level for the collaborator."}
    )

    def _validate_permission_level(self, permission_level: str):
        if not isinstance(permission_level, str):
            raise TypeError("permission_level must be a string.")

    user_id: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Optional user ID backing the collaborator (may be absent for service accounts)."
        },
    )

    def _validate_user_id(self, user_id: Optional[str]):
        if user_id is not None and not isinstance(user_id, str):
            raise TypeError("user_id must be a string.")



from typing import List, Optional, Tuple

import click
from rich import print as rprint
import tabulate

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand
from anyscale.resource_quota.models import CreateResourceQuota, ResourceQuota
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("resource-quota", help="Anyscale resource quota commands.")
def resource_quota_cli() -> None:
    pass


def _format_resource_quotas(resource_quotas: List[ResourceQuota]) -> str:
    table_rows = []
    for resource_quota in resource_quotas:
        table_rows.append(
            [
                resource_quota.id,
                resource_quota.name,
                resource_quota.cloud_id,
                resource_quota.project_id,
                resource_quota.user_id,
                resource_quota.is_enabled,
                resource_quota.created_at.strftime("%m/%d/%Y"),
                resource_quota.deleted_at.strftime("%m/%d/%Y")
                if resource_quota.deleted_at
                else None,
                resource_quota.quota,
            ]
        )
    table = tabulate.tabulate(
        table_rows,
        headers=[
            "ID",
            "NAME",
            "CLOUD ID",
            "PROJECT ID",
            "USER ID",
            "IS ENABLED",
            "CREATED AT",
            "DELETED AT",
            "QUOTA",
        ],
        tablefmt="plain",
    )

    return f"Resource quotas:\n{table}"


@resource_quota_cli.command(
    name="create",
    help="Create a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_CREATE_EXAMPLE,
)
@click.option(
    "-n", "--name", required=True, help="Name of the resource quota to create.",
)
@click.option(
    "--cloud",
    required=True,
    help="Name of the cloud that this resource quota applies to.",
)
@click.option(
    "--project",
    default=None,
    help="Name of the project that this resource quota applies to.",
)
@click.option(
    "--user-email",
    default=None,
    help="Email of the user that this resource quota applies to.",
)
@click.option(
    "--num-cpus",
    required=False,
    help="The quota limit for the number of CPUs.",
    type=int,
)
@click.option(
    "--num-instances",
    required=False,
    help="The quota limit for the number of instances.",
    type=int,
)
@click.option(
    "--num-gpus",
    required=False,
    help="The quota limit for the total number of GPUs.",
    type=int,
)
@click.option(
    "--num-accelerators",
    required=False,
    help="The quota limit for the number of accelerators. Example: --num-accelerators A100-80G 10",
    nargs=2,
    type=(str, int),
    multiple=True,
)
def create(  # noqa: PLR0913
    name: str,
    cloud: str,
    project: Optional[str],
    user_email: Optional[str],
    num_cpus: Optional[int],
    num_instances: Optional[int],
    num_gpus: Optional[int],
    num_accelerators: List[Tuple[str, int]],
) -> None:
    """Creates a resource quota.

    A name and cloud name must be provided.

    `$ anyscale resource-quota create -n my-resource-quota --cloud my-cloud --project my-project --user-email test@myorg.com --num-cpus 10 --num-instances 10 --num-gpus 10 --num-accelerators L4 5 --num-accelerators T4 10`
    """
    create_resource_quota = CreateResourceQuota(
        name=name,
        cloud=cloud,
        project=project,
        user_email=user_email,
        num_cpus=num_cpus,
        num_instances=num_instances,
        num_gpus=num_gpus,
        num_accelerators=dict(num_accelerators),
    )

    try:
        with log.spinner("Creating resource quota..."):
            resource_quota = anyscale.resource_quota.create(create_resource_quota)

        create_resource_quota_message = [f"Name: {name}\nCloud name: {cloud}"]
        if project:
            create_resource_quota_message.append(f"Project name: {project}")
        if user_email:
            create_resource_quota_message.append(f"User email: {user_email}")
        if num_cpus:
            create_resource_quota_message.append(f"Number of CPUs: {num_cpus}")
        if num_instances:
            create_resource_quota_message.append(
                f"Number of instances: {num_instances}"
            )
        if num_gpus:
            create_resource_quota_message.append(f"Number of GPUs: {num_gpus}")
        if num_accelerators:
            create_resource_quota_message.append(
                f"Number of accelerators: {dict(num_accelerators)}"
            )

        log.info("\n".join(create_resource_quota_message))
        log.info(f"Resource quota created successfully ID: {resource_quota.id}")

    except ValueError as e:
        log.error(f"Error creating resource quota: {e}")
        return


@resource_quota_cli.command(
    name="list",
    help="List resource quotas.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_LIST_EXAMPLE,
)
@click.option(
    "-n", "--name", required=False, help="The name filter for the resource quotas.",
)
@click.option(
    "--cloud", required=False, help="The cloud filter for the resource quotas.",
)
@click.option(
    "--creator-id",
    required=False,
    help="The creator ID filter for the resource quotas.",
)
@click.option(
    "--is-enabled",
    required=False,
    default=None,
    help="The is_enabled filter for the resource quotas.",
    type=bool,
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list_resource_quotas(
    name: Optional[str],
    cloud: Optional[str],
    creator_id: Optional[str],
    is_enabled: Optional[bool],
    max_items: int,
) -> None:
    """List resource quotas.

    `$ anyscale resource-quota list -n my-resource-quota --cloud my-cloud`
    """
    resource_quotas = anyscale.resource_quota.list(
        name=name,
        cloud=cloud,
        creator_id=creator_id,
        is_enabled=is_enabled,
        max_items=max_items,
    )

    rprint(_format_resource_quotas(resource_quotas))


@resource_quota_cli.command(
    name="delete",
    help="Delete a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_DELETE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to delete.",
)
def delete(id: str) -> None:  # noqa: A002
    """Deletes a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota delete --id rsq_123`
    """
    try:
        with log.spinner("Deleting resource quota..."):
            anyscale.resource_quota.delete(resource_quota_id=id)
    except ValueError as e:
        log.error(f"Error deleting resource quota: {e}")
        return

    log.info(f"Resource quota with ID {id} deleted successfully.")


@resource_quota_cli.command(
    name="enable",
    help="Enable a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_ENABLE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to enable.",
)
def enable(id: str) -> None:  # noqa: A002
    """Enables a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota enable --id rsq_123`
    """
    try:
        with log.spinner("Setting resource quota status..."):
            anyscale.resource_quota.enable(resource_quota_id=id)
    except ValueError as e:
        log.error(f"Error enabling resource quota: {e}")
        return

    log.info(f"Enabled resource quota with ID {id} successfully.")


@resource_quota_cli.command(
    name="disable",
    help="Disable a resource quota.",
    cls=AnyscaleCommand,
    is_beta=True,
    example=command_examples.RESOURCE_QUOTAS_DISABLE_EXAMPLE,
)
@click.option(
    "--id", required=True, help="ID of the resource quota to disable.",
)
def disable(id: str) -> None:  # noqa: A002
    """Disables a resource quota.

    An ID of resource quota must be provided.

    `$ anyscale resource-quota disable --id rsq_123`
    """
    try:
        with log.spinner("Setting resource quota status..."):
            anyscale.resource_quota.disable(resource_quota_id=id)
    except ValueError as e:
        log.error(f"Error disabling resource quota: {e}")
        return

    log.info(f"Disabled resource quota with ID {id} successfully.")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class TaskGroupedAggregateMetrics(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'total': 'int',
        'count_finished': 'int',
        'count_failed': 'int',
        'count_running': 'int',
        'count_pending_schedule': 'int',
        'count_pending_dep': 'int',
        'max_duration_ts_ns': 'int',
        'min_duration_ts_ns': 'int',
        'avg_duration_ts_ns': 'int',
        'max_execution_ns': 'int',
        'min_execution_ns': 'int',
        'avg_execution_ns': 'int'
    }

    attribute_map = {
        'total': 'total',
        'count_finished': 'count_finished',
        'count_failed': 'count_failed',
        'count_running': 'count_running',
        'count_pending_schedule': 'count_pending_schedule',
        'count_pending_dep': 'count_pending_dep',
        'max_duration_ts_ns': 'max_duration_ts_ns',
        'min_duration_ts_ns': 'min_duration_ts_ns',
        'avg_duration_ts_ns': 'avg_duration_ts_ns',
        'max_execution_ns': 'max_execution_ns',
        'min_execution_ns': 'min_execution_ns',
        'avg_execution_ns': 'avg_execution_ns'
    }

    def __init__(self, total=None, count_finished=None, count_failed=None, count_running=None, count_pending_schedule=None, count_pending_dep=None, max_duration_ts_ns=None, min_duration_ts_ns=None, avg_duration_ts_ns=None, max_execution_ns=None, min_execution_ns=None, avg_execution_ns=None, local_vars_configuration=None):  # noqa: E501
        """TaskGroupedAggregateMetrics - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._total = None
        self._count_finished = None
        self._count_failed = None
        self._count_running = None
        self._count_pending_schedule = None
        self._count_pending_dep = None
        self._max_duration_ts_ns = None
        self._min_duration_ts_ns = None
        self._avg_duration_ts_ns = None
        self._max_execution_ns = None
        self._min_execution_ns = None
        self._avg_execution_ns = None
        self.discriminator = None

        self.total = total
        self.count_finished = count_finished
        self.count_failed = count_failed
        self.count_running = count_running
        self.count_pending_schedule = count_pending_schedule
        self.count_pending_dep = count_pending_dep
        self.max_duration_ts_ns = max_duration_ts_ns
        self.min_duration_ts_ns = min_duration_ts_ns
        self.avg_duration_ts_ns = avg_duration_ts_ns
        if max_execution_ns is not None:
            self.max_execution_ns = max_execution_ns
        if min_execution_ns is not None:
            self.min_execution_ns = min_execution_ns
        if avg_execution_ns is not None:
            self.avg_execution_ns = avg_execution_ns

    @property
    def total(self):
        """Gets the total of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The total of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._total

    @total.setter
    def total(self, total):
        """Sets the total of this TaskGroupedAggregateMetrics.


        :param total: The total of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and total is None:  # noqa: E501
            raise ValueError("Invalid value for `total`, must not be `None`")  # noqa: E501

        self._total = total

    @property
    def count_finished(self):
        """Gets the count_finished of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The count_finished of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._count_finished

    @count_finished.setter
    def count_finished(self, count_finished):
        """Sets the count_finished of this TaskGroupedAggregateMetrics.


        :param count_finished: The count_finished of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and count_finished is None:  # noqa: E501
            raise ValueError("Invalid value for `count_finished`, must not be `None`")  # noqa: E501

        self._count_finished = count_finished

    @property
    def count_failed(self):
        """Gets the count_failed of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The count_failed of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._count_failed

    @count_failed.setter
    def count_failed(self, count_failed):
        """Sets the count_failed of this TaskGroupedAggregateMetrics.


        :param count_failed: The count_failed of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and count_failed is None:  # noqa: E501
            raise ValueError("Invalid value for `count_failed`, must not be `None`")  # noqa: E501

        self._count_failed = count_failed

    @property
    def count_running(self):
        """Gets the count_running of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The count_running of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._count_running

    @count_running.setter
    def count_running(self, count_running):
        """Sets the count_running of this TaskGroupedAggregateMetrics.


        :param count_running: The count_running of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and count_running is None:  # noqa: E501
            raise ValueError("Invalid value for `count_running`, must not be `None`")  # noqa: E501

        self._count_running = count_running

    @property
    def count_pending_schedule(self):
        """Gets the count_pending_schedule of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The count_pending_schedule of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._count_pending_schedule

    @count_pending_schedule.setter
    def count_pending_schedule(self, count_pending_schedule):
        """Sets the count_pending_schedule of this TaskGroupedAggregateMetrics.


        :param count_pending_schedule: The count_pending_schedule of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and count_pending_schedule is None:  # noqa: E501
            raise ValueError("Invalid value for `count_pending_schedule`, must not be `None`")  # noqa: E501

        self._count_pending_schedule = count_pending_schedule

    @property
    def count_pending_dep(self):
        """Gets the count_pending_dep of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The count_pending_dep of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._count_pending_dep

    @count_pending_dep.setter
    def count_pending_dep(self, count_pending_dep):
        """Sets the count_pending_dep of this TaskGroupedAggregateMetrics.


        :param count_pending_dep: The count_pending_dep of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and count_pending_dep is None:  # noqa: E501
            raise ValueError("Invalid value for `count_pending_dep`, must not be `None`")  # noqa: E501

        self._count_pending_dep = count_pending_dep

    @property
    def max_duration_ts_ns(self):
        """Gets the max_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The max_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._max_duration_ts_ns

    @max_duration_ts_ns.setter
    def max_duration_ts_ns(self, max_duration_ts_ns):
        """Sets the max_duration_ts_ns of this TaskGroupedAggregateMetrics.


        :param max_duration_ts_ns: The max_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and max_duration_ts_ns is None:  # noqa: E501
            raise ValueError("Invalid value for `max_duration_ts_ns`, must not be `None`")  # noqa: E501

        self._max_duration_ts_ns = max_duration_ts_ns

    @property
    def min_duration_ts_ns(self):
        """Gets the min_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The min_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._min_duration_ts_ns

    @min_duration_ts_ns.setter
    def min_duration_ts_ns(self, min_duration_ts_ns):
        """Sets the min_duration_ts_ns of this TaskGroupedAggregateMetrics.


        :param min_duration_ts_ns: The min_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and min_duration_ts_ns is None:  # noqa: E501
            raise ValueError("Invalid value for `min_duration_ts_ns`, must not be `None`")  # noqa: E501

        self._min_duration_ts_ns = min_duration_ts_ns

    @property
    def avg_duration_ts_ns(self):
        """Gets the avg_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The avg_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._avg_duration_ts_ns

    @avg_duration_ts_ns.setter
    def avg_duration_ts_ns(self, avg_duration_ts_ns):
        """Sets the avg_duration_ts_ns of this TaskGroupedAggregateMetrics.


        :param avg_duration_ts_ns: The avg_duration_ts_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and avg_duration_ts_ns is None:  # noqa: E501
            raise ValueError("Invalid value for `avg_duration_ts_ns`, must not be `None`")  # noqa: E501

        self._avg_duration_ts_ns = avg_duration_ts_ns

    @property
    def max_execution_ns(self):
        """Gets the max_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The max_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._max_execution_ns

    @max_execution_ns.setter
    def max_execution_ns(self, max_execution_ns):
        """Sets the max_execution_ns of this TaskGroupedAggregateMetrics.


        :param max_execution_ns: The max_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """

        self._max_execution_ns = max_execution_ns

    @property
    def min_execution_ns(self):
        """Gets the min_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The min_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._min_execution_ns

    @min_execution_ns.setter
    def min_execution_ns(self, min_execution_ns):
        """Sets the min_execution_ns of this TaskGroupedAggregateMetrics.


        :param min_execution_ns: The min_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """

        self._min_execution_ns = min_execution_ns

    @property
    def avg_execution_ns(self):
        """Gets the avg_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501


        :return: The avg_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :rtype: int
        """
        return self._avg_execution_ns

    @avg_execution_ns.setter
    def avg_execution_ns(self, avg_execution_ns):
        """Sets the avg_execution_ns of this TaskGroupedAggregateMetrics.


        :param avg_execution_ns: The avg_execution_ns of this TaskGroupedAggregateMetrics.  # noqa: E501
        :type: int
        """

        self._avg_execution_ns = avg_execution_ns

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TaskGroupedAggregateMetrics):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, TaskGroupedAggregateMetrics):
            return True

        return self.to_dict() != other.to_dict()
